import copy
from datetime import datetime, timezone
from unittest.mock import Mock, call, patch

import pytest
from arcade_tdk.errors import RetryableToolError, ToolExecutionError
from slack_sdk.errors import SlackApiError
from slack_sdk.web.async_slack_response import AsyncSlackResponse

from arcade_slack.constants import MAX_PAGINATION_SIZE_LIMIT
from arcade_slack.models import ConversationType, ConversationTypeSlackName
from arcade_slack.tools.chat import (
    get_channel_metadata_by_name,
    get_conversation_metadata_by_id,
    get_direct_message_conversation_metadata_by_username,
    get_members_in_channel_by_name,
    get_members_in_conversation_by_id,
    get_messages_in_channel_by_name,
    get_messages_in_conversation_by_id,
    get_messages_in_direct_message_conversation_by_username,
    get_messages_in_multi_person_dm_conversation_by_usernames,
    get_multi_person_dm_conversation_metadata_by_usernames,
    list_conversations_metadata,
    list_direct_message_conversations_metadata,
    list_group_direct_message_conversations_metadata,
    list_private_channels_metadata,
    list_public_channels_metadata,
    send_dm_to_user,
    send_message_to_channel,
)
from arcade_slack.utils import extract_basic_user_info, extract_conversation_metadata


@pytest.fixture
def mock_list_conversations_metadata(mocker):
    return mocker.patch("arcade_slack.tools.chat.list_conversations_metadata", autospec=True)


@pytest.fixture
def mock_channel_info() -> dict:
    return {"name": "general", "id": "C12345", "is_member": True, "is_channel": True}


@pytest.mark.asyncio
async def test_send_dm_to_user(mock_context, mock_chat_slack_client):
    mock_chat_slack_client.users_list.return_value = {
        "ok": True,
        "members": [{"name": "testuser", "id": "U12345"}],
    }
    mock_chat_slack_client.conversations_open.return_value = {
        "ok": True,
        "channel": {"id": "D12345"},
    }
    mock_slack_response = Mock(spec=AsyncSlackResponse)
    mock_slack_response.data = {"ok": True}
    mock_chat_slack_client.chat_postMessage.return_value = mock_slack_response

    response = await send_dm_to_user(mock_context, "testuser", "Hello!")

    assert response["response"]["ok"] is True
    mock_chat_slack_client.users_list.assert_called_once()
    mock_chat_slack_client.conversations_open.assert_called_once_with(users=["U12345"])
    mock_chat_slack_client.chat_postMessage.assert_called_once_with(channel="D12345", text="Hello!")


@pytest.mark.asyncio
async def test_send_dm_to_inexistent_user(mock_context, mock_chat_slack_client):
    mock_chat_slack_client.users_list.return_value = {
        "ok": True,
        "members": [{"name": "testuser", "id": "U12345"}],
    }

    with pytest.raises(RetryableToolError):
        await send_dm_to_user(mock_context, "inexistent_user", "Hello!")

    mock_chat_slack_client.users_list.assert_called_once()
    mock_chat_slack_client.conversations_open.assert_not_called()
    mock_chat_slack_client.chat_postMessage.assert_not_called()


@pytest.mark.asyncio
async def test_send_message_to_channel(mock_context, mock_chat_slack_client):
    mock_chat_slack_client.conversations_list.return_value = {
        "ok": True,
        "channels": [{"id": "C12345", "name": "general", "is_member": True, "is_group": True}],
    }
    mock_slack_response = Mock(spec=AsyncSlackResponse)
    mock_slack_response.data = {"ok": True}
    mock_chat_slack_client.chat_postMessage.return_value = mock_slack_response

    response = await send_message_to_channel(mock_context, "general", "Hello, channel!")

    assert response["response"]["ok"] is True
    mock_chat_slack_client.conversations_list.assert_called_once()
    mock_chat_slack_client.chat_postMessage.assert_called_once_with(
        channel="C12345", text="Hello, channel!"
    )


@pytest.mark.asyncio
async def test_send_message_to_inexistent_channel(mock_context, mock_chat_slack_client):
    mock_chat_slack_client.conversations_list.return_value = {
        "ok": True,
        "channels": [],
    }

    with pytest.raises(RetryableToolError):
        await send_message_to_channel(mock_context, "inexistent_channel", "Hello!")

    mock_chat_slack_client.conversations_list.assert_called_once()
    mock_chat_slack_client.chat_postMessage.assert_not_called()


@pytest.mark.asyncio
async def test_list_conversations_metadata_with_default_args(
    mock_context, mock_chat_slack_client, mock_channel_info
):
    mock_chat_slack_client.conversations_list.return_value = {
        "ok": True,
        "channels": [mock_channel_info],
    }

    response = await list_conversations_metadata(mock_context)

    assert response["conversations"] == [extract_conversation_metadata(mock_channel_info)]
    assert response["next_cursor"] is None

    mock_chat_slack_client.conversations_list.assert_called_once_with(
        types=",".join([conv_type.value for conv_type in ConversationTypeSlackName]),
        exclude_archived=True,
        limit=MAX_PAGINATION_SIZE_LIMIT,
        cursor=None,
    )


@pytest.mark.asyncio
async def test_list_conversations_metadata_filtering_single_conversation_type(
    mock_context, mock_chat_slack_client, mock_channel_info
):
    mock_chat_slack_client.conversations_list.return_value = {
        "ok": True,
        "channels": [mock_channel_info],
    }

    response = await list_conversations_metadata(
        mock_context, conversation_types=ConversationType.PUBLIC_CHANNEL
    )

    assert response["conversations"] == [extract_conversation_metadata(mock_channel_info)]
    assert response["next_cursor"] is None

    mock_chat_slack_client.conversations_list.assert_called_once_with(
        types=ConversationTypeSlackName.PUBLIC_CHANNEL.value,
        exclude_archived=True,
        limit=MAX_PAGINATION_SIZE_LIMIT,
        cursor=None,
    )


@pytest.mark.asyncio
async def test_list_conversations_metadata_filtering_multiple_conversation_types(
    mock_context, mock_chat_slack_client, mock_channel_info
):
    mock_chat_slack_client.conversations_list.return_value = {
        "ok": True,
        "channels": [mock_channel_info],
    }

    response = await list_conversations_metadata(
        mock_context,
        conversation_types=[
            ConversationTypeSlackName.PUBLIC_CHANNEL,
            ConversationTypeSlackName.PRIVATE_CHANNEL,
        ],
    )

    assert response["conversations"] == [extract_conversation_metadata(mock_channel_info)]
    assert response["next_cursor"] is None

    mock_chat_slack_client.conversations_list.assert_called_once_with(
        types=f"{ConversationTypeSlackName.PUBLIC_CHANNEL.value},{ConversationTypeSlackName.PRIVATE_CHANNEL.value}",
        exclude_archived=True,
        limit=MAX_PAGINATION_SIZE_LIMIT,
        cursor=None,
    )


@pytest.mark.asyncio
async def test_list_conversations_metadata_with_custom_pagination_args(
    mock_context, mock_chat_slack_client, mock_channel_info
):
    mock_chat_slack_client.conversations_list.return_value = {
        "ok": True,
        "channels": [mock_channel_info] * 3,
        "response_metadata": {"next_cursor": "456"},
    }

    response = await list_conversations_metadata(mock_context, limit=3, next_cursor="123")

    assert response["conversations"] == [
        extract_conversation_metadata(mock_channel_info) for _ in range(3)
    ]
    assert response["next_cursor"] == "456"

    mock_chat_slack_client.conversations_list.assert_called_once_with(
        types=",".join([conv_type.value for conv_type in ConversationTypeSlackName]),
        exclude_archived=True,
        limit=3,
        cursor="123",
    )


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "faulty_slack_function_name, tool_function, tool_args",
    [
        ("users_list", send_dm_to_user, ("testuser", "Hello!")),
        ("conversations_list", send_message_to_channel, ("general", "Hello!")),
    ],
)
async def test_tools_with_slack_error(
    mock_context, mock_chat_slack_client, faulty_slack_function_name, tool_function, tool_args
):
    getattr(mock_chat_slack_client, faulty_slack_function_name).side_effect = SlackApiError(
        message="test_slack_error",
        response={"ok": False, "error": "test_slack_error"},
    )

    with pytest.raises(ToolExecutionError) as e:
        await tool_function(mock_context, *tool_args)
        assert "test_slack_error" in str(e.value)


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "tool_function, conversation_type",
    [
        (list_public_channels_metadata, ConversationType.PUBLIC_CHANNEL),
        (list_private_channels_metadata, ConversationType.PRIVATE_CHANNEL),
        (
            list_group_direct_message_conversations_metadata,
            ConversationType.MULTI_PERSON_DIRECT_MESSAGE,
        ),
        (list_direct_message_conversations_metadata, ConversationType.DIRECT_MESSAGE),
    ],
)
async def test_list_channels_metadata(
    mock_context,
    mock_list_conversations_metadata,
    tool_function,
    conversation_type,
):
    response = await tool_function(mock_context, limit=3)

    mock_list_conversations_metadata.assert_called_once_with(
        mock_context, conversation_types=[conversation_type], limit=3
    )

    assert response == mock_list_conversations_metadata.return_value


@pytest.mark.asyncio
async def test_get_conversation_metadata_by_id(
    mock_context, mock_chat_slack_client, mock_channel_info
):
    mock_chat_slack_client.conversations_info.return_value = {
        "ok": True,
        "channel": mock_channel_info,
    }

    response = await get_conversation_metadata_by_id(mock_context, "C12345")

    assert response == extract_conversation_metadata(mock_channel_info)
    mock_chat_slack_client.conversations_info.assert_called_once_with(
        channel="C12345",
        include_locale=True,
        include_num_members=True,
    )


@pytest.mark.asyncio
@patch("arcade_slack.tools.chat.list_conversations_metadata")
async def test_get_conversation_metadata_by_id_slack_api_error(
    mock_list_conversations_metadata, mock_context, mock_chat_slack_client, mock_channel_info
):
    mock_channel_info["name"] = "whatever_conversation_should_be_present_in_additional_prompt"
    mock_list_conversations_metadata.return_value = {
        "conversations": [extract_conversation_metadata(mock_channel_info)],
        "response_metadata": {"next_cursor": None},
    }
    mock_chat_slack_client.conversations_info.side_effect = SlackApiError(
        message="channel_not_found",
        response={"ok": False, "error": "channel_not_found"},
    )

    with pytest.raises(RetryableToolError) as e:
        await get_conversation_metadata_by_id(mock_context, "C12345")

        assert (
            "whatever_conversation_should_be_present_in_additional_prompt"
            in e.additional_prompt_content
        )

    mock_chat_slack_client.conversations_info.assert_called_once_with(
        channel="C12345",
        include_locale=True,
        include_num_members=True,
    )
    mock_list_conversations_metadata.assert_called_once_with(mock_context)


@pytest.mark.asyncio
async def test_get_conversation_metadata_by_name(
    mock_context, mock_list_conversations_metadata, mock_channel_info
):
    sample_conversation = extract_conversation_metadata(mock_channel_info)
    mock_list_conversations_metadata.return_value = {
        "conversations": [sample_conversation],
        "next_cursor": None,
    }

    response = await get_channel_metadata_by_name(mock_context, sample_conversation["name"])

    assert response == sample_conversation
    mock_list_conversations_metadata.assert_called_once_with(
        context=mock_context,
        conversation_types=[
            ConversationType.PUBLIC_CHANNEL,
            ConversationType.PRIVATE_CHANNEL,
        ],
        next_cursor=None,
    )


@pytest.mark.asyncio
async def test_get_channel_metadata_by_name_triggering_pagination(
    mock_context, mock_list_conversations_metadata, mock_channel_info
):
    target_channel = extract_conversation_metadata(mock_channel_info)
    another_channel = extract_conversation_metadata(mock_channel_info)
    another_channel["name"] = "another_channel"

    mock_list_conversations_metadata.side_effect = [
        {
            "conversations": [another_channel],
            "next_cursor": "123",
        },
        {
            "conversations": [target_channel],
            "next_cursor": None,
        },
    ]

    response = await get_channel_metadata_by_name(mock_context, target_channel["name"])

    assert response == target_channel
    assert mock_list_conversations_metadata.call_count == 2
    mock_list_conversations_metadata.assert_has_calls([
        call(
            context=mock_context,
            conversation_types=[ConversationType.PUBLIC_CHANNEL, ConversationType.PRIVATE_CHANNEL],
            next_cursor=None,
        ),
        call(
            context=mock_context,
            conversation_types=[ConversationType.PUBLIC_CHANNEL, ConversationType.PRIVATE_CHANNEL],
            next_cursor="123",
        ),
    ])


@pytest.mark.asyncio
async def test_get_channel_metadata_by_name_not_found(
    mock_context, mock_list_conversations_metadata, mock_channel_info
):
    first_channel = extract_conversation_metadata(mock_channel_info)
    second_channel = extract_conversation_metadata(mock_channel_info)
    second_channel["name"] = "second_channel"

    mock_list_conversations_metadata.side_effect = [
        {
            "conversations": [second_channel],
            "next_cursor": "123",
        },
        {
            "conversations": [first_channel],
            "next_cursor": None,
        },
    ]

    with pytest.raises(RetryableToolError):
        await get_channel_metadata_by_name(mock_context, "inexistent_channel")

    assert mock_list_conversations_metadata.call_count == 2
    mock_list_conversations_metadata.assert_has_calls([
        call(
            context=mock_context,
            conversation_types=[ConversationType.PUBLIC_CHANNEL, ConversationType.PRIVATE_CHANNEL],
            next_cursor=None,
        ),
        call(
            context=mock_context,
            conversation_types=[ConversationType.PUBLIC_CHANNEL, ConversationType.PRIVATE_CHANNEL],
            next_cursor="123",
        ),
    ])


@pytest.mark.asyncio
@patch("arcade_slack.tools.chat.async_paginate")
@patch("arcade_slack.tools.chat.get_user_info_by_id")
async def test_get_members_from_conversation_id(
    mock_get_user_info_by_id, mock_async_paginate, mock_context, mock_chat_slack_client
):
    member1 = {"id": "U123", "name": "testuser123"}
    member1_info = extract_basic_user_info(member1)
    member2 = {"id": "U456", "name": "testuser456"}
    member2_info = extract_basic_user_info(member2)

    mock_async_paginate.return_value = [member1["id"], member2["id"]], "token123"
    mock_get_user_info_by_id.side_effect = [member1_info, member2_info]

    response = await get_members_in_conversation_by_id(
        mock_context, conversation_id="C12345", limit=2
    )

    assert response == {
        "members": [member1_info, member2_info],
        "next_cursor": "token123",
    }
    mock_async_paginate.assert_called_once_with(
        mock_chat_slack_client.conversations_members,
        "members",
        limit=2,
        next_cursor=None,
        channel="C12345",
    )
    mock_get_user_info_by_id.assert_has_calls([
        call(mock_context, member1["id"]),
        call(mock_context, member2["id"]),
    ])


@pytest.mark.asyncio
@patch("arcade_slack.tools.chat.async_paginate")
@patch("arcade_slack.tools.chat.get_user_info_by_id")
@patch("arcade_slack.tools.chat.list_conversations_metadata")
async def test_get_members_from_conversation_id_channel_not_found(
    mock_list_conversations_metadata,
    mock_get_user_info_by_id,
    mock_async_paginate,
    mock_context,
    mock_chat_slack_client,
    mock_channel_info,
):
    conversations = [extract_conversation_metadata(mock_channel_info)] * 2
    mock_list_conversations_metadata.return_value = {
        "conversations": conversations,
        "next_cursor": None,
    }

    member1 = {"id": "U123", "name": "testuser123"}
    member1_info = extract_basic_user_info(member1)
    member2 = {"id": "U456", "name": "testuser456"}
    member2_info = extract_basic_user_info(member2)

    mock_async_paginate.side_effect = SlackApiError(
        message="channel_not_found",
        response={"ok": False, "error": "channel_not_found"},
    )
    mock_get_user_info_by_id.side_effect = [member1_info, member2_info]

    with pytest.raises(RetryableToolError):
        await get_members_in_conversation_by_id(mock_context, conversation_id="C12345", limit=2)

    mock_async_paginate.assert_called_once_with(
        mock_chat_slack_client.conversations_members,
        "members",
        limit=2,
        next_cursor=None,
        channel="C12345",
    )
    mock_get_user_info_by_id.assert_not_called()


@pytest.mark.asyncio
@patch("arcade_slack.tools.chat.list_conversations_metadata")
@patch("arcade_slack.tools.chat.get_members_in_conversation_by_id")
async def test_get_members_in_channel_by_name(
    mock_get_members_in_conversation_by_id,
    mock_list_conversations_metadata,
    mock_context,
    mock_channel_info,
):
    mock_list_conversations_metadata.return_value = {
        "conversations": [extract_conversation_metadata(mock_channel_info)],
        "next_cursor": None,
    }

    response = await get_members_in_channel_by_name(
        mock_context, mock_channel_info["name"], limit=2
    )

    assert response == mock_get_members_in_conversation_by_id.return_value
    mock_list_conversations_metadata.assert_called_once_with(
        context=mock_context,
        conversation_types=[
            ConversationType.PUBLIC_CHANNEL,
            ConversationType.PRIVATE_CHANNEL,
        ],
        next_cursor=None,
    )
    mock_get_members_in_conversation_by_id.assert_called_once_with(
        context=mock_context,
        conversation_id="C12345",
        limit=2,
        next_cursor=None,
    )


@pytest.mark.asyncio
@patch("arcade_slack.tools.chat.list_conversations_metadata")
@patch("arcade_slack.tools.chat.get_members_in_conversation_by_id")
async def test_get_members_in_channel_by_name_triggering_pagination(
    mock_get_members_in_conversation_by_id,
    mock_list_conversations_metadata,
    mock_context,
    mock_channel_info,
):
    conversation1 = copy.deepcopy(mock_channel_info)
    conversation1["name"] = "conversation1"
    conversation2 = copy.deepcopy(mock_channel_info)
    conversation2["name"] = "conversation2"

    mock_list_conversations_metadata.side_effect = [
        {
            "conversations": [extract_conversation_metadata(conversation1)],
            "next_cursor": "123",
        },
        {
            "conversations": [extract_conversation_metadata(conversation2)],
            "next_cursor": None,
        },
    ]

    response = await get_members_in_channel_by_name(mock_context, conversation2["name"], limit=2)

    assert response == mock_get_members_in_conversation_by_id.return_value
    mock_list_conversations_metadata.assert_has_calls([
        call(
            context=mock_context,
            conversation_types=[ConversationType.PUBLIC_CHANNEL, ConversationType.PRIVATE_CHANNEL],
            next_cursor=None,
        ),
        call(
            context=mock_context,
            conversation_types=[ConversationType.PUBLIC_CHANNEL, ConversationType.PRIVATE_CHANNEL],
            next_cursor="123",
        ),
    ])
    mock_get_members_in_conversation_by_id.assert_called_once_with(
        context=mock_context,
        conversation_id="C12345",
        limit=2,
        next_cursor=None,
    )


@pytest.mark.asyncio
async def test_get_conversation_history_by_id(mock_context, mock_chat_slack_client):
    mock_chat_slack_client.conversations_history.return_value = {
        "ok": True,
        "messages": [{"text": "Hello, world!"}],
    }

    response = await get_messages_in_conversation_by_id(mock_context, "C12345", limit=1)

    assert response == {"messages": [{"text": "Hello, world!"}], "next_cursor": None}
    mock_chat_slack_client.conversations_history.assert_called_once_with(
        channel="C12345",
        include_all_metadata=True,
        inclusive=True,
        limit=1,
        cursor=None,
    )


# TODO: pass a current unix timestamp to the tool, instead of mocking the datetime
# conversion. Have to wait until arcade.core.annotations.Inferrable is implemented.
@pytest.mark.asyncio
@patch("arcade_slack.tools.chat.convert_relative_datetime_to_unix_timestamp")
@patch("arcade_slack.tools.chat.datetime")
async def test_get_conversation_history_by_id_with_relative_datetime_args(
    mock_datetime,
    mock_convert_relative_datetime_to_unix_timestamp,
    mock_context,
    mock_chat_slack_client,
):
    mock_chat_slack_client.conversations_history.return_value = {
        "ok": True,
        "messages": [{"text": "Hello, world!"}],
    }

    expected_oldest_timestamp = 1716489600
    expected_latest_timestamp = 1716403200

    # Ideally we'd pass the current unix timestamp to the function, instead of mocking, but
    # currently there's no way to have a tool argument that is not exposed to the LLM. We
    # should have that soon, though.
    mock_datetime.now.return_value = datetime(2025, 1, 1, 0, 0, 0, tzinfo=timezone.utc)
    expected_current_unix_timestamp = int(mock_datetime.now.return_value.timestamp())
    mock_convert_relative_datetime_to_unix_timestamp.side_effect = [
        expected_latest_timestamp,
        expected_oldest_timestamp,
    ]

    response = await get_messages_in_conversation_by_id(
        mock_context, "C12345", oldest_relative="02:00:00", latest_relative="01:00:00", limit=1
    )

    assert response == {"messages": [{"text": "Hello, world!"}], "next_cursor": None}
    mock_convert_relative_datetime_to_unix_timestamp.assert_has_calls([
        call("01:00:00", expected_current_unix_timestamp),
        call("02:00:00", expected_current_unix_timestamp),
    ])
    mock_chat_slack_client.conversations_history.assert_called_once_with(
        channel="C12345",
        include_all_metadata=True,
        inclusive=True,
        limit=1,
        cursor=None,
        oldest=expected_oldest_timestamp,
        latest=expected_latest_timestamp,
    )


# TODO: pass a current unix timestamp to the tool, instead of mocking the datetime
# conversion. Have to wait until arcade.core.annotations.Inferrable is implemented.
@pytest.mark.asyncio
@patch("arcade_slack.tools.chat.convert_datetime_to_unix_timestamp")
async def test_get_conversation_history_by_id_with_absolute_datetime_args(
    mock_convert_datetime_to_unix_timestamp, mock_context, mock_chat_slack_client
):
    mock_chat_slack_client.conversations_history.return_value = {
        "ok": True,
        "messages": [{"text": "Hello, world!"}],
    }

    expected_latest_timestamp = 1716403200
    expected_oldest_timestamp = 1716489600

    # Ideally we'd pass the current unix timestamp to the function, instead of mocking, but
    # currently there's no way to have a tool argument that is not exposed to the LLM. We
    # should have that soon, though.
    mock_convert_datetime_to_unix_timestamp.side_effect = [
        expected_latest_timestamp,
        expected_oldest_timestamp,
    ]

    response = await get_messages_in_conversation_by_id(
        mock_context,
        "C12345",
        oldest_datetime="2025-01-01 00:00:00",
        latest_datetime="2025-01-02 00:00:00",
        limit=1,
    )

    assert response == {"messages": [{"text": "Hello, world!"}], "next_cursor": None}
    mock_convert_datetime_to_unix_timestamp.assert_has_calls([
        call("2025-01-02 00:00:00"),
        call("2025-01-01 00:00:00"),
    ])
    mock_chat_slack_client.conversations_history.assert_called_once_with(
        channel="C12345",
        include_all_metadata=True,
        inclusive=True,
        limit=1,
        cursor=None,
        oldest=expected_oldest_timestamp,
        latest=expected_latest_timestamp,
    )


@pytest.mark.asyncio
async def test_get_conversation_history_by_id_with_messed_oldest_args(
    mock_context, mock_chat_slack_client
):
    with pytest.raises(ToolExecutionError):
        await get_messages_in_conversation_by_id(
            mock_context,
            "C12345",
            oldest_datetime="2025-01-01 00:00:00",
            oldest_relative="01:00:00",
        )


@pytest.mark.asyncio
async def test_get_conversation_history_by_id_with_messed_latest_args(
    mock_context, mock_chat_slack_client
):
    with pytest.raises(ToolExecutionError):
        await get_messages_in_conversation_by_id(
            mock_context,
            "C12345",
            latest_datetime="2025-01-01 00:00:00",
            latest_relative="01:00:00",
        )


@pytest.mark.asyncio
async def test_get_conversation_history_by_name(mock_context, mock_chat_slack_client):
    mock_chat_slack_client.conversations_list.return_value = {
        "ok": True,
        "channels": [
            {
                "id": "C12345",
                "name": "general",
                "is_member": True,
                "is_channel": True,
            }
        ],
    }
    mock_chat_slack_client.conversations_history.return_value = {
        "ok": True,
        "messages": [{"text": "Hello, world!"}],
    }

    response = await get_messages_in_channel_by_name(mock_context, "general", limit=1)

    assert response == {"messages": [{"text": "Hello, world!"}], "next_cursor": None}
    mock_chat_slack_client.conversations_history.assert_called_once_with(
        channel="C12345", include_all_metadata=True, inclusive=True, limit=1, cursor=None
    )


@pytest.mark.asyncio
@patch("arcade_slack.tools.chat.retrieve_conversations_by_user_ids")
async def test_get_direct_message_conversation_metadata_by_username(
    mock_retrieve_conversations_by_user_ids,
    mock_context,
    mock_chat_slack_client,
    mock_users_slack_client,
):
    mock_chat_slack_client.auth_test.return_value = {
        "ok": True,
        "user_id": "U1",
        "team_id": "T1",
        "user": "user1",
    }

    mock_users_slack_client.users_list.return_value = {
        "ok": True,
        "members": [
            {"id": "U1", "name": "user1"},
            {"id": "U2", "name": "user2"},
        ],
        "response_metadata": {"next_cursor": None},
    }

    conversation = {
        "id": "C12345",
        "type": ConversationTypeSlackName.IM.value,
        "is_im": True,
        "members": ["U1", "U2"],
    }

    mock_retrieve_conversations_by_user_ids.return_value = [conversation]

    response = await get_direct_message_conversation_metadata_by_username(
        context=mock_context, username="user2"
    )

    assert response == conversation
    mock_retrieve_conversations_by_user_ids.assert_called_once_with(
        list_conversations_func=list_conversations_metadata,
        get_members_in_conversation_func=get_members_in_conversation_by_id,
        context=mock_context,
        conversation_types=[ConversationType.DIRECT_MESSAGE],
        user_ids=["U1", "U2"],
        exact_match=True,
        limit=1,
        next_cursor=None,
    )


@pytest.mark.asyncio
@patch("arcade_slack.tools.chat.retrieve_conversations_by_user_ids")
async def test_get_direct_message_conversation_metadata_by_username_username_not_found(
    mock_retrieve_conversations_by_user_ids,
    mock_context,
    mock_chat_slack_client,
    mock_users_slack_client,
):
    mock_chat_slack_client.users_identity.return_value = {
        "ok": True,
        "user": {"id": "U1", "name": "user1"},
        "team": {"id": "T1", "name": "team1"},
    }

    mock_users_slack_client.users_list.return_value = {
        "ok": True,
        "members": [
            {"id": "U1", "name": "user1"},
            {"id": "U2", "name": "user2"},
        ],
        "response_metadata": {"next_cursor": None},
    }

    mock_retrieve_conversations_by_user_ids.side_effect = TimeoutError()

    with pytest.raises(RetryableToolError):
        await get_direct_message_conversation_metadata_by_username(
            context=mock_context, username="user999"
        )


@pytest.mark.asyncio
@patch("arcade_slack.tools.chat.get_messages_in_conversation_by_id")
@patch("arcade_slack.tools.chat.get_direct_message_conversation_metadata_by_username")
async def test_get_messages_in_direct_conversation_by_username(
    mock_get_direct_message_conversation_metadata_by_username,
    mock_get_messages_in_conversation_by_id,
    mock_context,
):
    mock_get_direct_message_conversation_metadata_by_username.return_value = {
        "id": "C12345",
    }

    response = await get_messages_in_direct_message_conversation_by_username(
        context=mock_context, username="user2"
    )

    assert response == mock_get_messages_in_conversation_by_id.return_value
    mock_get_direct_message_conversation_metadata_by_username.assert_called_once_with(
        context=mock_context, username="user2"
    )
    mock_get_messages_in_conversation_by_id.assert_called_once_with(
        context=mock_context,
        conversation_id="C12345",
        oldest_relative=None,
        latest_relative=None,
        oldest_datetime=None,
        latest_datetime=None,
        limit=None,
        next_cursor=None,
    )


@pytest.mark.asyncio
@patch("arcade_slack.tools.chat.get_direct_message_conversation_metadata_by_username")
async def test_get_messages_in_direct_conversation_by_username_not_found(
    mock_get_direct_message_conversation_metadata_by_username,
    mock_context,
):
    mock_get_direct_message_conversation_metadata_by_username.return_value = None

    with pytest.raises(ToolExecutionError):
        await get_messages_in_direct_message_conversation_by_username(
            context=mock_context, username="user2"
        )


@pytest.mark.asyncio
@patch("arcade_slack.tools.chat.retrieve_conversations_by_user_ids")
async def test_get_multi_person_direct_message_conversation_metadata_by_username(
    mock_retrieve_conversations_by_user_ids,
    mock_context,
    mock_chat_slack_client,
    mock_users_slack_client,
):
    mock_chat_slack_client.auth_test.return_value = {
        "ok": True,
        "user_id": "U1",
        "team_id": "T1",
        "user": "user1",
    }

    mock_users_slack_client.users_list.return_value = {
        "ok": True,
        "members": [
            {"id": "U1", "name": "user1"},
            {"id": "U2", "name": "user2"},
            {"id": "U3", "name": "user3"},
            {"id": "U4", "name": "user4"},
            {"id": "U5", "name": "user5"},
        ],
        "response_metadata": {"next_cursor": None},
    }

    conversation = {
        "id": "C12345",
        "type": ConversationTypeSlackName.MPIM.value,
        "is_mpim": True,
        "members": ["U1", "U4", "U5"],
    }

    mock_retrieve_conversations_by_user_ids.return_value = [conversation]

    response = await get_multi_person_dm_conversation_metadata_by_usernames(
        context=mock_context, usernames=["user1", "user4", "user5"]
    )

    assert response == conversation
    mock_retrieve_conversations_by_user_ids.assert_called_once_with(
        list_conversations_func=list_conversations_metadata,
        get_members_in_conversation_func=get_members_in_conversation_by_id,
        context=mock_context,
        conversation_types=[ConversationType.MULTI_PERSON_DIRECT_MESSAGE],
        user_ids=["U1", "U4", "U5"],
        exact_match=True,
        limit=1,
        next_cursor=None,
    )


@pytest.mark.asyncio
@patch("arcade_slack.tools.chat.retrieve_conversations_by_user_ids")
async def test_get_multi_person_direct_message_conversation_metadata_by_username_username_not_found(
    mock_retrieve_conversations_by_user_ids,
    mock_context,
    mock_chat_slack_client,
    mock_users_slack_client,
):
    mock_chat_slack_client.users_identity.return_value = {
        "ok": True,
        "user": {"id": "U1", "name": "user1"},
        "team": {"id": "T1", "name": "team1"},
    }

    mock_users_slack_client.users_list.return_value = {
        "ok": True,
        "members": [
            {"id": "U1", "name": "user1"},
            {"id": "U2", "name": "user2"},
        ],
        "response_metadata": {"next_cursor": None},
    }

    mock_retrieve_conversations_by_user_ids.side_effect = TimeoutError()

    with pytest.raises(RetryableToolError):
        await get_multi_person_dm_conversation_metadata_by_usernames(
            context=mock_context, usernames=["user999", "user1", "user2"]
        )


@pytest.mark.asyncio
@patch("arcade_slack.tools.chat.get_messages_in_conversation_by_id")
@patch("arcade_slack.tools.chat.get_multi_person_dm_conversation_metadata_by_usernames")
async def test_get_messages_in_multi_person_dm_conversation_by_usernames(
    mock_get_multi_person_dm_conversation_metadata_by_usernames,
    mock_get_messages_in_conversation_by_id,
    mock_context,
):
    mock_get_multi_person_dm_conversation_metadata_by_usernames.return_value = {
        "id": "C12345",
    }

    response = await get_messages_in_multi_person_dm_conversation_by_usernames(
        context=mock_context, usernames=["user1", "user4", "user5"]
    )

    assert response == mock_get_messages_in_conversation_by_id.return_value

    mock_get_multi_person_dm_conversation_metadata_by_usernames.assert_called_once_with(
        context=mock_context, usernames=["user1", "user4", "user5"]
    )

    mock_get_messages_in_conversation_by_id.assert_called_once_with(
        context=mock_context,
        conversation_id="C12345",
        oldest_relative=None,
        latest_relative=None,
        oldest_datetime=None,
        latest_datetime=None,
        limit=None,
        next_cursor=None,
    )


@pytest.mark.asyncio
@patch("arcade_slack.tools.chat.get_multi_person_dm_conversation_metadata_by_usernames")
async def test_get_messages_in_multi_person_dm_conversation_by_usernames_not_found(
    mock_get_multi_person_dm_conversation_metadata_by_usernames,
    mock_context,
):
    mock_get_multi_person_dm_conversation_metadata_by_usernames.return_value = None

    with pytest.raises(ToolExecutionError):
        await get_messages_in_direct_message_conversation_by_username(
            context=mock_context, username="user2"
        )



class SlackToolkitError(Exception):
    """Base class for all Slack toolkit errors."""


class PaginationTimeoutError(SlackToolkitError):
    """Raised when a timeout occurs during pagination."""

    def __init__(self, timeout_seconds: int):
        self.timeout_seconds = timeout_seconds
        super().__init__(f"The pagination process timed out after {timeout_seconds} seconds.")


class ItemNotFoundError(SlackToolkitError):
    """Raised when an item is not found."""


class UsernameNotFoundError(SlackToolkitError):
    """Raised when a user is not found by the username searched"""

    def __init__(self, usernames_found: list[str], username_not_found: str) -> None:
        self.usernames_found = usernames_found
        self.username_not_found = username_not_found


class ConversationNotFoundError(SlackToolkitError):
    """Raised when a conversation is not found"""


class DirectMessageConversationNotFoundError(ConversationNotFoundError):
    """Raised when a direct message conversation searched is not found"""



from unittest.mock import patch

import pytest
from arcade_tdk.errors import RetryableToolError
from slack_sdk.errors import SlackApiError

from arcade_slack.tools.users import get_user_info_by_id, list_users
from arcade_slack.utils import extract_basic_user_info


@pytest.fixture
def mock_slack_client(mocker):
    mock_client = mocker.patch("arcade_slack.tools.users.AsyncWebClient", autospec=True)
    return mock_client.return_value


@pytest.mark.asyncio
async def test_get_user_info_by_id_success(mock_context, mock_slack_client):
    # Mock the response from slackClient.users_info
    mock_user = {
        "id": "U12345",
        "name": "testuser",
        "real_name": "Test User",
        "profile": {"email": "testuser@example.com"},
    }
    mock_slack_client.users_info.return_value = {"ok": True, "user": mock_user}

    # Call the function
    response = await get_user_info_by_id(mock_context, user_id="U12345")

    # Verify that the correct Slack API method was called
    mock_slack_client.users_info.assert_called_once_with(user="U12345")

    # Verify the response
    expected_response = extract_basic_user_info(mock_user)
    assert response == expected_response


@pytest.mark.asyncio
@patch("arcade_slack.tools.users.list_users")
async def test_get_user_info_by_id_user_not_found(mock_list_users, mock_context, mock_slack_client):
    error_response = {"ok": False, "error": "user_not_found"}
    mock_slack_client.users_info.side_effect = SlackApiError(
        message="User not found",
        response=error_response,
    )

    existing_user = {"id": "U12345", "name": "testuser"}
    mock_list_users.return_value = {"users": [existing_user]}

    with pytest.raises(RetryableToolError) as e:
        await get_user_info_by_id(mock_context, user_id="U99999")

        assert existing_user["id"] in e.value.additional_prompt_content
        assert existing_user["name"] in e.value.additional_prompt_content

    mock_slack_client.users_info.assert_called_once_with(user="U99999")
    mock_list_users.assert_called_once_with(mock_context)


@pytest.mark.asyncio
async def test_list_users_success(mock_context, mock_slack_client):
    mock_slack_client.users_list.return_value = {"ok": True, "members": [{"id": "U12345"}]}
    response = await list_users(mock_context)
    assert response == {
        "users": [extract_basic_user_info({"id": "U12345"})],
        "next_cursor": None,
    }


@pytest.mark.asyncio
async def test_list_users_with_pagination_success(mock_context, mock_slack_client):
    mock_slack_client.users_list.side_effect = [
        {
            "ok": True,
            "members": [{"id": "U12345"}],
            "response_metadata": {"next_cursor": "cursor_xyz"},
        },
        {
            "ok": True,
            "members": [{"id": "U123456"}],
            "response_metadata": {"next_cursor": None},
        },
    ]
    response = await list_users(mock_context, limit=3)
    assert response == {
        "users": [
            extract_basic_user_info({"id": "U12345"}),
            extract_basic_user_info({"id": "U123456"}),
        ],
        "next_cursor": None,
    }



import pytest
from arcade_tdk import ToolAuthorizationContext, ToolContext


@pytest.fixture
def mock_context():
    mock_auth = ToolAuthorizationContext(token="fake-token")  # noqa: S106
    return ToolContext(authorization=mock_auth)


@pytest.fixture
def mock_chat_slack_client(mocker):
    mock_client = mocker.patch("arcade_slack.tools.chat.AsyncWebClient", autospec=True)
    return mock_client.return_value


@pytest.fixture
def mock_users_slack_client(mocker):
    mock_client = mocker.patch("arcade_slack.tools.users.AsyncWebClient", autospec=True)
    return mock_client.return_value



from typing import Annotated, Any, cast

from arcade_tdk import ToolContext, tool
from arcade_tdk.auth import Slack
from arcade_tdk.errors import RetryableToolError
from slack_sdk.errors import SlackApiError
from slack_sdk.web.async_client import AsyncWebClient

from arcade_slack.constants import MAX_PAGINATION_TIMEOUT_SECONDS
from arcade_slack.models import SlackPaginationNextCursor, SlackUser
from arcade_slack.utils import (
    async_paginate,
    extract_basic_user_info,
    is_user_a_bot,
    is_user_deleted,
)


@tool(
    requires_auth=Slack(
        scopes=["users:read", "users:read.email"],
    )
)
async def get_user_info_by_id(
    context: ToolContext,
    user_id: Annotated[str, "The ID of the user to get"],
) -> Annotated[dict[str, Any], "The user's information"]:
    """Get the information of a user in Slack."""

    token = (
        context.authorization.token if context.authorization and context.authorization.token else ""
    )
    slackClient = AsyncWebClient(token=token)

    try:
        response = await slackClient.users_info(user=user_id)
    except SlackApiError as e:
        if e.response.get("error") == "user_not_found":
            users = await list_users(context)
            available_users = ", ".join(f"{user['id']} ({user['name']})" for user in users["users"])

            raise RetryableToolError(
                "User not found",
                developer_message=f"User with ID '{user_id}' not found.",
                additional_prompt_content=f"Available users: {available_users}",
                retry_after_ms=500,
            )

    user_dict_raw: dict[str, Any] = response.get("user", {}) or {}
    user_dict = cast(SlackUser, user_dict_raw)
    user = SlackUser(**user_dict)
    return dict(**extract_basic_user_info(user))


@tool(
    requires_auth=Slack(
        scopes=["users:read", "users:read.email"],
    )
)
async def list_users(
    context: ToolContext,
    exclude_bots: Annotated[bool | None, "Whether to exclude bots from the results"] = True,
    limit: Annotated[int | None, "The maximum number of users to return."] = None,
    next_cursor: Annotated[str | None, "The next cursor token to use for pagination."] = None,
) -> Annotated[dict, "The users' info"]:
    """List all users in the authenticated user's Slack team."""

    token = (
        context.authorization.token if context.authorization and context.authorization.token else ""
    )
    slackClient = AsyncWebClient(token=token)

    users, next_cursor = await async_paginate(
        func=slackClient.users_list,
        response_key="members",
        limit=limit,
        next_cursor=cast(SlackPaginationNextCursor, next_cursor),
        max_pagination_timeout_seconds=MAX_PAGINATION_TIMEOUT_SECONDS,
    )

    users = [
        extract_basic_user_info(user)
        for user in users
        if not is_user_deleted(user) and (not exclude_bots or not is_user_a_bot(user))
    ]

    return {"users": users, "next_cursor": next_cursor}



import json
from datetime import timedelta

from arcade_evals import (
    BinaryCritic,
    DatetimeCritic,
    EvalRubric,
    EvalSuite,
    ExpectedToolCall,
    SimilarityCritic,
    tool_eval,
)
from arcade_tdk import ToolCatalog

import arcade_slack
from arcade_slack.critics import RelativeTimeBinaryCritic
from arcade_slack.tools.chat import (
    get_channel_metadata_by_name,
    get_conversation_metadata_by_id,
    get_direct_message_conversation_metadata_by_username,
    get_members_in_channel_by_name,
    get_members_in_conversation_by_id,
    get_messages_in_channel_by_name,
    get_messages_in_conversation_by_id,
    get_messages_in_direct_message_conversation_by_username,
    get_messages_in_multi_person_dm_conversation_by_usernames,
    list_conversations_metadata,
    list_direct_message_conversations_metadata,
    list_group_direct_message_conversations_metadata,
    list_private_channels_metadata,
    list_public_channels_metadata,
    send_dm_to_user,
    send_message_to_channel,
)

# Evaluation rubric
rubric = EvalRubric(
    fail_threshold=0.8,
    warn_threshold=0.9,
)


catalog = ToolCatalog()
# Register the Slack tools
catalog.add_module(arcade_slack)


@tool_eval()
def send_message_eval_suite() -> EvalSuite:
    """Create an evaluation suite for Slack messaging tools."""
    suite = EvalSuite(
        name="Slack Messaging Tools Evaluation",
        system_message="You are an AI assistant that can send direct messages and post messages to channels in Slack using the provided tools.",
        catalog=catalog,
        rubric=rubric,
    )

    # Send DM to User Scenarios
    suite.add_case(
        name="Send DM to user with clear username",
        user_message="Send a direct message to johndoe saying 'Hello, can we meet at 3 PM?'",
        expected_tool_calls=[
            ExpectedToolCall(
                func=send_dm_to_user,
                args={
                    "user_name": "johndoe",
                    "message": "Hello, can we meet at 3 PM?",
                },
            )
        ],
        critics=[
            BinaryCritic(critic_field="user_name", weight=0.5),
            SimilarityCritic(critic_field="message", weight=0.5, similarity_threshold=0.9),
        ],
    )

    suite.add_case(
        name="Send DM with ambiguous username",
        user_message="ask him for an update on the project",
        expected_tool_calls=[
            ExpectedToolCall(
                func=send_dm_to_user,
                args={
                    "user_name": "john",
                    "message": "Hi John, could you please provide an update on the Acme project?",
                },
            )
        ],
        critics=[
            BinaryCritic(critic_field="user_name", weight=0.75),
            SimilarityCritic(critic_field="message", weight=0.25, similarity_threshold=0.6),
        ],
        additional_messages=[
            {"role": "user", "content": "Message John about the Acme project deadline"},
            {
                "role": "assistant",
                "content": "",
                "tool_calls": [
                    {
                        "id": "call_1",
                        "type": "function",
                        "function": {
                            "name": "Slack_ListUsers",
                            "arguments": '{"exclude_bots":true}',
                        },
                    }
                ],
            },
            {
                "role": "tool",
                "content": json.dumps({
                    "users": [
                        {
                            "display_name": "john",
                            "email": "john@randomtech.com",
                            "id": "abc123",
                            "is_bot": False,
                            "name": "john",
                            "real_name": "John Doe",
                            "timezone": "America/Los_Angeles",
                        },
                        {
                            "display_name": "jack",
                            "email": "jack@randomtech.com",
                            "id": "def456",
                            "is_bot": False,
                            "name": "jack",
                            "real_name": "Jack Doe",
                            "timezone": "America/Los_Angeles",
                        },
                    ]
                }),
                "tool_call_id": "call_1",
                "name": "Slack_ListUsers",
            },
            {
                "role": "assistant",
                "content": "What would you like to include in the message to John about the Acme project deadline?",
            },
        ],
    )

    suite.add_case(
        name="Send DM with username in different format",
        user_message="yes, send it",
        expected_tool_calls=[
            ExpectedToolCall(
                func=send_dm_to_user,
                args={
                    "user_name": "jane.doe",
                    "message": "Hi Jane, I need to reschedule our meeting. When are you available?",
                },
            )
        ],
        critics=[
            BinaryCritic(critic_field="user_name", weight=0.75),
            SimilarityCritic(critic_field="message", weight=0.25, similarity_threshold=0.6),
        ],
        additional_messages=[
            {"role": "user", "content": "Message Jane.Doe asking to reschedule our meeting"},
            {
                "role": "assistant",
                "content": "",
                "tool_calls": [
                    {
                        "id": "call_1",
                        "type": "function",
                        "function": {
                            "name": "Slack_ListUsers",
                            "arguments": '{"exclude_bots":true}',
                        },
                    }
                ],
            },
            {
                "role": "tool",
                "content": json.dumps({
                    "users": [
                        {
                            "display_name": "jane.doe",
                            "email": "jane@randomtech.com",
                            "id": "abc123",
                            "is_bot": False,
                            "name": "jane.doe",
                            "real_name": "Jane Doe",
                            "timezone": "America/Los_Angeles",
                        },
                        {
                            "display_name": "jack",
                            "email": "jack@randomtech.com",
                            "id": "def456",
                            "is_bot": False,
                            "name": "jack",
                            "real_name": "Jack Doe",
                            "timezone": "America/Los_Angeles",
                        },
                    ]
                }),
                "tool_call_id": "call_1",
                "name": "Slack_ListUsers",
            },
            {
                "role": "assistant",
                "content": "I found a user with the name 'jane.doe'. Would you like to send a message to them?",
            },
        ],
    )

    # Send Message to Channel Scenarios
    suite.add_case(
        name="Send message to channel with clear name",
        user_message="Post 'The new feature is now live!' in the #announcements channel",
        expected_tool_calls=[
            ExpectedToolCall(
                func=send_message_to_channel,
                args={
                    "channel_name": "announcements",
                    "message": "The new feature is now live!",
                },
            )
        ],
        critics=[
            BinaryCritic(critic_field="channel_name", weight=0.5),
            SimilarityCritic(critic_field="message", weight=0.5),
        ],
    )

    suite.add_case(
        name="Send message to channel with ambiguous name",
        user_message="Inform the team in the general channel about the upcoming maintenance",
        expected_tool_calls=[
            ExpectedToolCall(
                func=send_message_to_channel,
                args={
                    "channel_name": "general",
                    "message": "Attention team: There will be upcoming maintenance. Please save your work and expect some downtime.",
                },
            )
        ],
        critics=[
            SimilarityCritic(critic_field="channel_name", weight=0.8),
            SimilarityCritic(critic_field="message", weight=0.2, similarity_threshold=0.6),
        ],
    )

    # Adversarial Scenarios
    suite.add_case(
        name="Ambiguous between DM and channel message",
        user_message="general",
        expected_tool_calls=[
            ExpectedToolCall(
                func=send_message_to_channel,
                args={
                    "channel_name": "general",
                    "message": "Great job on the presentation!",
                },
            )
        ],
        critics=[
            SimilarityCritic(critic_field="channel_name", weight=0.4),
            SimilarityCritic(critic_field="message", weight=0.6),
        ],
        additional_messages=[
            {"role": "user", "content": "Send 'Great job on the presentation!' to the team"},
            {
                "role": "assistant",
                "content": "",
                "tool_calls": [
                    {
                        "id": "call_1",
                        "type": "function",
                        "function": {
                            "name": "Slack_ListPublicChannelsMetadata",
                            "arguments": '{"limit":20}',
                        },
                    }
                ],
            },
            {
                "role": "tool",
                "content": json.dumps({
                    "conversations": [
                        {
                            "conversation_type": "public_channel",
                            "id": "channel1",
                            "is_archived": False,
                            "is_member": True,
                            "is_private": False,
                            "name": "random",
                            "num_members": 999,
                            "purpose": "Random stuff",
                        },
                        {
                            "conversation_type": "public_channel",
                            "id": "channel2",
                            "is_archived": False,
                            "is_member": True,
                            "is_private": False,
                            "name": "general",
                            "num_members": 999,
                            "purpose": "Just a general channel",
                        },
                    ],
                    "next_cursor": "",
                }),
                "tool_call_id": "call_1",
                "name": "Slack_ListPublicChannelsMetadata",
            },
            {
                "role": "assistant",
                "content": 'To send the message "Great job on the presentation!" to the team, please let me know which Slack channel you\'d like to use:\n\n1. #random\n2. #general\n\nPlease let me know your choice!',
            },
        ],
    )

    # Multiple recipients in DM request
    suite.add_case(
        name="Multiple recipients in DM request",
        user_message="Send DMs to the users 'alice' and 'bob' about pushing the meeting tomorrow. I have to much work to do.",
        expected_tool_calls=[
            ExpectedToolCall(
                func=send_dm_to_user,
                args={
                    "user_name": "alice",
                    "message": "Hi Alice, about our meeting tomorrow, let's reschedule? I am swamped with work.",
                },
            ),
            ExpectedToolCall(
                func=send_dm_to_user,
                args={
                    "user_name": "bob",
                    "message": "Hi Bob, about our meeting tomorrow, let's reschedule? I am swamped with work.",
                },
            ),
        ],
        critics=[
            SimilarityCritic(critic_field="user_name", weight=0.75),
            SimilarityCritic(critic_field="message", weight=0.25, similarity_threshold=0.5),
        ],
    )

    suite.add_case(
        name="Channel name similar to username",
        user_message="Post 'sounds great!' in john-project channel",
        expected_tool_calls=[
            ExpectedToolCall(
                func=send_message_to_channel,
                args={
                    "channel_name": "john-project",
                    "message": "Sounds great!",
                },
            )
        ],
        critics=[
            BinaryCritic(critic_field="channel_name", weight=0.5),
            SimilarityCritic(critic_field="message", weight=0.5),
        ],
    )

    return suite


@tool_eval()
def list_conversations_eval_suite() -> EvalSuite:
    """Create an evaluation suite for tools listing conversations."""
    suite = EvalSuite(
        name="Slack Messaging Tools Evaluation",
        system_message="You are an AI assistant that can interact with Slack to send messages and get information from conversations, users, etc.",
        catalog=catalog,
        rubric=rubric,
    )

    cases = [
        (
            "List my conversations",
            "List all conversations I am a member of",
            list_conversations_metadata,
        ),
        (
            "List public channels",
            "List all public channels",
            list_public_channels_metadata,
        ),
        (
            "List private channels",
            "List all private channels",
            list_private_channels_metadata,
        ),
        (
            "List group direct message channels",
            "List all group direct message channels",
            list_group_direct_message_conversations_metadata,
        ),
        (
            "List individual direct message channels",
            "List all individual direct message channels",
            list_direct_message_conversations_metadata,
        ),
        (
            "List direct message channels",
            "List all direct message channels",
            list_direct_message_conversations_metadata,
        ),
        (
            "List public and private channels",
            "List public and private channels I am a member of",
            list_public_channels_metadata,
            list_private_channels_metadata,
        ),
        (
            "List public channels and direct message conversations",
            "List public channels and direct message conversations I am a member of",
            list_public_channels_metadata,
            list_direct_message_conversations_metadata,
        ),
    ]

    for name, user_message, *expect_called_tool_functions in cases:
        suite.add_case(
            name=name,
            user_message=user_message,
            expected_tool_calls=[
                ExpectedToolCall(
                    func=tool_function,
                    args={},
                )
                for tool_function in expect_called_tool_functions
            ],
        )

    return suite


@tool_eval()
def get_conversations_metadata_eval_suite() -> EvalSuite:
    """Create an evaluation suite for tools getting conversations metadata."""
    suite = EvalSuite(
        name="Slack Tools Evaluation",
        system_message="You are an AI assistant that can interact with Slack to get information from conversations, users, etc.",
        catalog=catalog,
        rubric=rubric,
    )

    suite.add_case(
        name="Get channel metadata by name",
        user_message="Get the metadata of the #general channel",
        expected_tool_calls=[
            ExpectedToolCall(
                func=get_channel_metadata_by_name,
                args={
                    "channel_name": "general",
                },
            ),
        ],
        critics=[
            BinaryCritic(critic_field="conversation_name", weight=1.0),
        ],
    )

    suite.add_case(
        name="Get conversation metadata by id",
        user_message="Get the metadata of the conversation with id '1234567890'",
        expected_tool_calls=[
            ExpectedToolCall(
                func=get_conversation_metadata_by_id,
                args={
                    "conversation_id": "1234567890",
                },
            ),
        ],
        critics=[
            BinaryCritic(critic_field="conversation_id", weight=1.0),
        ],
    )

    get_metadata_by_username_user_messages = [
        "get the metadata of the direct message conversation with the user 'jane.doe'"
        "get data about my private conversation with the user 'jane.doe'",
        "get data about my IM conversation with the 'jane.doe'",
    ]

    for i, user_message in enumerate(get_metadata_by_username_user_messages):
        suite.add_case(
            name=f"Get direct message conversation metadata by username {i}",
            user_message=user_message,
            expected_tool_calls=[
                ExpectedToolCall(
                    func=get_direct_message_conversation_metadata_by_username,
                    args={
                        "username": "jane.doe",
                    },
                ),
            ],
        )

    return suite


@tool_eval()
def get_conversations_members_eval_suite() -> EvalSuite:
    """Create an evaluation suite for tools getting conversations members."""
    suite = EvalSuite(
        name="Slack Tools Evaluation",
        system_message="You are an AI assistant that can interact with Slack to send messages and get information from conversations, users, etc.",
        catalog=catalog,
        rubric=rubric,
    )

    user_messages = [
        "Get the members of the #general channel",
        "Get the members of the general channel",
        "Get a list of people in the #general channel",
        "Get a list of people in the general channel",
        "Show me who's in the #general channel",
        "Show me who's in the general channel",
        "Who is in the #general channel?",
        "Who is in the general channel?",
    ]

    for user_message in user_messages:
        suite.add_case(
            name=f"Get channel members by name: {user_message}",
            user_message=user_message,
            expected_tool_calls=[
                ExpectedToolCall(
                    func=get_members_in_channel_by_name,
                    args={
                        "channel_name": "general",
                    },
                ),
            ],
            critics=[
                BinaryCritic(critic_field="conversation_name", weight=1.0),
            ],
        )

    suite.add_case(
        name="Get conversation members by id",
        user_message="Get the members of the conversation with id '1234567890'",
        expected_tool_calls=[
            ExpectedToolCall(
                func=get_members_in_conversation_by_id,
                args={
                    "conversation_id": "1234567890",
                },
            ),
        ],
        critics=[
            BinaryCritic(critic_field="conversation_id", weight=1.0),
        ],
    )

    return suite


@tool_eval()
def get_messages_in_channel_eval_suite() -> EvalSuite:
    """Create an evaluation suite for tools getting messages in channels."""
    suite = EvalSuite(
        name="Slack Chat Tools Evaluation",
        system_message="You are an AI assistant that can interact with Slack to send messages and get information from conversations, users, etc.",
        catalog=catalog,
        rubric=rubric,
    )

    no_arguments_user_messages_by_channel_name = [
        "what are the latest messages in the #general channel",
        "show me the messages in the general channel",
        "list the messages in the #general channel",
        "list the messages in the general channel",
    ]

    for i, user_message in enumerate(no_arguments_user_messages_by_channel_name):
        suite.add_case(
            name=f"Get messages in conversation by name {i}: '{user_message}'",
            user_message=user_message,
            expected_tool_calls=[
                ExpectedToolCall(
                    func=get_messages_in_channel_by_name,
                    args={
                        "conversation_name": "general",
                    },
                ),
            ],
            critics=[
                BinaryCritic(critic_field="conversation_name", weight=1.0),
            ],
        )

    no_arguments_user_messages_by_conversation_id = [
        "Get the history of the conversation with id '1234567890'",
        "Get the history of the conversation with id '1234567890'",
        "list the messages in the conversation with id '1234567890'",
        "list the messages in the conversation with id '1234567890'",
    ]

    for user_message in no_arguments_user_messages_by_conversation_id:
        suite.add_case(
            name=f"Get conversation history by id: '{user_message}'",
            user_message=user_message,
            expected_tool_calls=[
                ExpectedToolCall(
                    func=get_messages_in_conversation_by_id,
                    args={
                        "conversation_id": "1234567890",
                    },
                ),
            ],
            critics=[
                BinaryCritic(critic_field="conversation_id", weight=1.0),
            ],
        )

    suite.add_case(
        name="Get conversation history with limit by name",
        user_message="Get the last 10 messages in the #general channel",
        expected_tool_calls=[
            ExpectedToolCall(
                func=get_messages_in_channel_by_name,
                args={
                    "conversation_name": "general",
                    "limit": 10,
                },
            ),
        ],
        critics=[
            BinaryCritic(critic_field="conversation_name", weight=1.0),
        ],
    )

    suite.add_case(
        name="Get conversation history with limit by id",
        user_message="Get the last 25 messages in the conversation with id '1234567890'",
        expected_tool_calls=[
            ExpectedToolCall(
                func=get_messages_in_conversation_by_id,
                args={
                    "conversation_id": "1234567890",
                    "limit": 25,
                },
            ),
        ],
        critics=[
            BinaryCritic(critic_field="conversation_id", weight=1.0),
        ],
    )

    # Relative time eval cases by id

    suite.add_case(
        name="Get conversation history oldest relative by id (2 days ago)",
        user_message="Get the messages in the conversation with id '1234567890' starting 2 days ago",
        expected_tool_calls=[
            ExpectedToolCall(
                func=get_messages_in_conversation_by_id,
                args={
                    "conversation_id": "1234567890",
                    "oldest_relative": "02:00:00",
                },
            ),
        ],
        critics=[
            BinaryCritic(critic_field="conversation_id", weight=0.5),
            RelativeTimeBinaryCritic(critic_field="oldest_relative", weight=0.5),
        ],
    )

    suite.add_case(
        name="Get conversation history oldest and latest relative by id",
        user_message="Get the messages in the conversation with id '1234567890' from 2 days ago to 3 hours ago",
        expected_tool_calls=[
            ExpectedToolCall(
                func=get_messages_in_conversation_by_id,
                args={
                    "conversation_id": "1234567890",
                    "oldest_relative": "02:00:00",
                    "latest_relative": "00:03:00",
                },
            ),
        ],
        critics=[
            BinaryCritic(critic_field="conversation_id", weight=1 / 3),
            RelativeTimeBinaryCritic(critic_field="oldest_relative", weight=1 / 3),
            RelativeTimeBinaryCritic(critic_field="latest_relative", weight=1 / 3),
        ],
    )

    suite.add_case(
        name="Get conversation history oldest relative by id (1 week ago)",
        user_message="Get the messages in the conversation with id '1234567890' starting 1 week ago",
        expected_tool_calls=[
            ExpectedToolCall(
                func=get_messages_in_conversation_by_id,
                args={
                    "conversation_id": "1234567890",
                    "oldest_relative": "07:00:00",
                },
            ),
        ],
        critics=[
            BinaryCritic(critic_field="conversation_id", weight=0.5),
            RelativeTimeBinaryCritic(critic_field="oldest_relative", weight=0.5),
        ],
    )

    suite.add_case(
        name="Get conversation history oldest relative by id (yesterday)",
        user_message="Get the messages in the conversation with id '1234567890' from yesterday",
        expected_tool_calls=[
            ExpectedToolCall(
                func=get_messages_in_conversation_by_id,
                args={
                    "conversation_id": "1234567890",
                    "oldest_relative": "01:00:00",
                },
            ),
        ],
        critics=[
            BinaryCritic(critic_field="conversation_id", weight=0.5),
            RelativeTimeBinaryCritic(critic_field="oldest_relative", weight=0.5),
        ],
    )

    # Relative time eval cases by name

    suite.add_case(
        name="Get conversation history oldest relative by name (2 days ago)",
        user_message="Get the messages in the #general channel starting 2 days ago",
        expected_tool_calls=[
            ExpectedToolCall(
                func=get_messages_in_channel_by_name,
                args={
                    "conversation_name": "general",
                    "oldest_relative": "02:00:00",
                },
            ),
        ],
        critics=[
            BinaryCritic(critic_field="conversation_name", weight=0.5),
            RelativeTimeBinaryCritic(critic_field="oldest_relative", weight=0.5),
        ],
    )

    suite.add_case(
        name="Get conversation history oldest and latest relative by name",
        user_message="Get the messages in the #general channel from 2 days ago to 3 hours ago",
        expected_tool_calls=[
            ExpectedToolCall(
                func=get_messages_in_channel_by_name,
                args={
                    "conversation_name": "general",
                    "oldest_relative": "02:00:00",
                    "latest_relative": "00:03:00",
                },
            ),
        ],
        critics=[
            BinaryCritic(critic_field="conversation_name", weight=1 / 3),
            RelativeTimeBinaryCritic(critic_field="oldest_relative", weight=1 / 3),
            RelativeTimeBinaryCritic(critic_field="latest_relative", weight=1 / 3),
        ],
    )

    suite.add_case(
        name="Get conversation history oldest relative by name (yesterday)",
        user_message="Get the messages in the #general channel from yesterday",
        expected_tool_calls=[
            ExpectedToolCall(
                func=get_messages_in_channel_by_name,
                args={
                    "conversation_name": "general",
                    "oldest_relative": "01:00:00",
                },
            ),
        ],
        critics=[
            BinaryCritic(critic_field="conversation_name", weight=0.5),
            RelativeTimeBinaryCritic(critic_field="oldest_relative", weight=0.5),
        ],
    )

    suite.add_case(
        name="Get conversation history oldest relative by name (last week)",
        user_message="Get the messages in the #general channel from last week",
        expected_tool_calls=[
            ExpectedToolCall(
                func=get_messages_in_channel_by_name,
                args={
                    "conversation_name": "general",
                    "oldest_relative": "07:00:00",
                },
            ),
        ],
        critics=[
            BinaryCritic(critic_field="conversation_name", weight=0.5),
            RelativeTimeBinaryCritic(critic_field="oldest_relative", weight=0.5),
        ],
    )

    # Absolute time eval cases by id

    suite.add_case(
        name="Get conversation history oldest absolute by id (on a specific date)",
        user_message="Get the messages in the conversation with id '1234567890' from 2025-01-20",
        expected_tool_calls=[
            ExpectedToolCall(
                func=get_messages_in_conversation_by_id,
                args={
                    "conversation_id": "1234567890",
                    "oldest_datetime": "2025-01-20 00:00:00",
                    "latest_datetime": "2025-01-20 23:59:59",
                },
            ),
        ],
        critics=[
            BinaryCritic(critic_field="conversation_id", weight=1 / 3),
            DatetimeCritic(
                critic_field="oldest_datetime", weight=1 / 3, max_difference=timedelta(minutes=2)
            ),
            DatetimeCritic(
                critic_field="latest_datetime", weight=1 / 3, max_difference=timedelta(minutes=2)
            ),
        ],
    )

    suite.add_case(
        name="Get conversation history oldest absolute by id (between a date range)",
        user_message="Get the messages in the conversation with id '1234567890' from 2025-01-20 to 2025-01-25",
        expected_tool_calls=[
            ExpectedToolCall(
                func=get_messages_in_conversation_by_id,
                args={
                    "conversation_id": "1234567890",
                    "oldest_datetime": "2025-01-20 00:00:00",
                    "latest_datetime": "2025-01-25 23:59:59",
                },
            ),
        ],
        critics=[
            BinaryCritic(critic_field="conversation_id", weight=1 / 3),
            DatetimeCritic(
                critic_field="oldest_datetime", weight=1 / 3, max_difference=timedelta(minutes=2)
            ),
            DatetimeCritic(
                critic_field="latest_datetime", weight=1 / 3, max_difference=timedelta(minutes=2)
            ),
        ],
    )

    suite.add_case(
        name="Get conversation history oldest absolute by name (on a specific date)",
        user_message="Get the messages in the #general channel from 2025-01-20",
        expected_tool_calls=[
            ExpectedToolCall(
                func=get_messages_in_channel_by_name,
                args={
                    "conversation_name": "general",
                    "oldest_datetime": "2025-01-20 00:00:00",
                    "latest_datetime": "2025-01-20 23:59:59",
                },
            ),
        ],
        critics=[
            BinaryCritic(critic_field="conversation_name", weight=1 / 3),
            # We use a timedelta of 10 seconds because sometimes the LLM will select the limit
            # date at 23:59:59, other times it'll select the next day at 00:00:00.
            DatetimeCritic(
                critic_field="oldest_datetime", weight=1 / 3, max_difference=timedelta(seconds=10)
            ),
            DatetimeCritic(
                critic_field="latest_datetime", weight=1 / 3, max_difference=timedelta(seconds=10)
            ),
        ],
    )

    suite.add_case(
        name="Get conversation history oldest absolute by name (between a date range)",
        user_message="Get the messages in the #general channel from 2025-01-20 to 2025-01-25",
        expected_tool_calls=[
            ExpectedToolCall(
                func=get_messages_in_channel_by_name,
                args={
                    "conversation_name": "general",
                    "oldest_datetime": "2025-01-20 00:00:00",
                    "latest_datetime": "2025-01-25 23:59:59",
                },
            ),
        ],
        critics=[
            BinaryCritic(critic_field="conversation_name", weight=1 / 3),
            # We use a timedelta of 10 seconds because sometimes the LLM will select the limit
            # date at 23:59:59, other times it'll select the next day at 00:00:00.
            DatetimeCritic(
                critic_field="oldest_datetime", weight=1 / 3, max_difference=timedelta(seconds=10)
            ),
            DatetimeCritic(
                critic_field="latest_datetime", weight=1 / 3, max_difference=timedelta(seconds=10)
            ),
        ],
    )

    # Eval case for pagination

    suite.add_case(
        name="Get conversation history with pagination",
        user_message="get the next 5 messages",
        expected_tool_calls=[
            ExpectedToolCall(
                func=get_messages_in_channel_by_name,
                args={
                    "conversation_name": "general",
                    "limit": 5,
                    "cursor": "dXNlcjpVsDjzOTZGVDlQRA==",
                },
            ),
        ],
        critics=[
            BinaryCritic(critic_field="conversation_name", weight=1 / 3),
            BinaryCritic(critic_field="cursor", weight=1 / 3),
            BinaryCritic(critic_field="limit", weight=1 / 3),
        ],
        additional_messages=[
            {"role": "user", "content": "Get the last 2 messages on the general channel"},
            {
                "role": "assistant",
                "content": "",
                "tool_calls": [
                    {
                        "id": "call_1",
                        "type": "function",
                        "function": {
                            "name": "Slack_GetConversationHistoryByName",
                            "arguments": json.dumps({
                                "conversation_name": "general",
                                "limit": 2,
                            }),
                        },
                    }
                ],
            },
            {
                "role": "tool",
                "content": json.dumps({
                    "messages": [
                        {
                            "blocks": [
                                {
                                    "block_id": "abc123",
                                    "elements": [
                                        {
                                            "elements": [
                                                {
                                                    "text": "Almost there, Boss, need to get some evals in!",
                                                    "type": "text",
                                                }
                                            ],
                                            "type": "rich_text_section",
                                        }
                                    ],
                                    "type": "rich_text",
                                }
                            ],
                            "client_msg_id": "msg_id_0",
                            "datetime_timestamp": "2025-01-21 16:59:55",
                            "team": "617263616465207465616D20697320617420626F7373206C6576656C",
                            "text": "Almost there, Boss, need to get some evals in!",
                            "ts": "1737507595.598529",
                            "type": "message",
                            "user": "77686F2069732074686520626F73733F",
                        },
                        {
                            "blocks": [
                                {
                                    "block_id": "xyz456",
                                    "elements": [
                                        {
                                            "elements": [
                                                {
                                                    "text": "hey, are the Slack Tools ready yet?",
                                                    "type": "text",
                                                }
                                            ],
                                            "type": "rich_text_section",
                                        }
                                    ],
                                    "type": "rich_text",
                                }
                            ],
                            "client_msg_id": "msg_id_1",
                            "datetime_timestamp": "2025-01-21 16:57:35",
                            "team": "617263616465207465616D20697320617420626F7373206C6576656C",
                            "text": "hey, are the Slack Tools ready yet?",
                            "ts": "1737507595.598529",
                            "type": "message",
                            "user": "73616D2069732074686520626F7373",
                        },
                    ],
                    "next_cursor": "dXNlcjpVsDjzOTZGVDlQRA==",
                }),
                "tool_call_id": "call_1",
                "name": "Slack_GetConversationHistoryByName",
            },
            {
                "role": "assistant",
                "content": 'Here are the last 2 messages from the general channel:\n\n1. **User:** 77686F2069732074686520626F73733F  \n   **Message:** "Almost there, Boss, need to get some evals in!"  \n   **Timestamp:** 2025-01-21 16:59:55\n\n2. **User:** 73616D2069732074686520626F7373  \n   **Message:** "hey, are the Slack Tools ready yet?"  \n   **Timestamp:** 2025-01-21 16:57:35',
            },
        ],
    )

    return suite


@tool_eval()
def get_messages_in_direct_message_eval_suite() -> EvalSuite:
    """Create an evaluation suite for tools getting messages in direct messages."""
    suite = EvalSuite(
        name="Slack Chat Tools Evaluation",
        system_message="You are an AI assistant that can interact with Slack to send messages and get information from conversations, users, etc.",
        catalog=catalog,
        rubric=rubric,
    )

    no_arguments_user_messages_by_username = [
        "what are the latest messages I exchanged with jane.doe",
        "show my messages with jane.doe on Slack",
        "list the messages I exchanged with jane.doe",
        "list the message history with jane.doe",
    ]

    for i, user_message in enumerate(no_arguments_user_messages_by_username):
        suite.add_case(
            name=f"{user_message} [{i}]",
            user_message=user_message,
            expected_tool_calls=[
                ExpectedToolCall(
                    func=get_messages_in_direct_message_conversation_by_username,
                    args={
                        "username": "jane.doe",
                    },
                ),
            ],
            critics=[
                BinaryCritic(critic_field="username", weight=1.0),
            ],
        )

    suite.add_case(
        name="get messages in direct conversation by username (on a specific date)",
        user_message="get the messages I exchanged with jane.doe on 2025-01-31",
        expected_tool_calls=[
            ExpectedToolCall(
                func=get_messages_in_direct_message_conversation_by_username,
                args={
                    "username": "jane.doe",
                    "oldest_datetime": "2025-01-31 00:00:00",
                    "latest_datetime": "2025-01-31 23:59:59",
                },
            ),
        ],
        critics=[
            BinaryCritic(critic_field="username", weight=1 / 3),
            DatetimeCritic(
                critic_field="oldest_datetime", weight=1 / 3, max_difference=timedelta(minutes=2)
            ),
            DatetimeCritic(
                critic_field="latest_datetime", weight=1 / 3, max_difference=timedelta(minutes=2)
            ),
        ],
    )

    suite.add_case(
        name="Get conversation history oldest relative by username (2 days ago)",
        user_message="Get the messages I exchanged with jane.doe starting 2 days ago",
        expected_tool_calls=[
            ExpectedToolCall(
                func=get_messages_in_direct_message_conversation_by_username,
                args={
                    "username": "jane.doe",
                    "oldest_relative": "02:00:00",
                },
            ),
        ],
        critics=[
            BinaryCritic(critic_field="username", weight=0.5),
            RelativeTimeBinaryCritic(critic_field="oldest_relative", weight=0.5),
        ],
    )

    return suite


@tool_eval()
def get_messages_in_multi_person_direct_message_eval_suite() -> EvalSuite:
    """Create an evaluation suite for tools getting messages in multi-person direct messages."""
    suite = EvalSuite(
        name="Slack Chat Tools Evaluation",
        system_message="You are an AI assistant that can interact with Slack to send messages and get information from conversations, users, etc.",
        catalog=catalog,
        rubric=rubric,
    )

    no_arguments_user_messages_by_username = [
        "what are the latest messages I exchanged together with the usernames john, ryan, and jennifer",
        "show the messages in the multi person dm with the usernames john, ryan, and jennifer on Slack",
        "list the messages I exchanged together with the usernames john, ryan, and jennifer",
        "list the message history together with the usernames john, ryan, and jennifer",
    ]

    for i, user_message in enumerate(no_arguments_user_messages_by_username):
        suite.add_case(
            name=f"{user_message} [{i}]",
            user_message=user_message,
            expected_tool_calls=[
                ExpectedToolCall(
                    func=get_messages_in_multi_person_dm_conversation_by_usernames,
                    args={
                        "usernames": ["john", "ryan", "jennifer"],
                    },
                ),
            ],
            critics=[
                BinaryCritic(critic_field="usernames", weight=1.0),
            ],
        )

    suite.add_case(
        name="get messages in direct conversation by username (on a specific date)",
        user_message="get the messages I exchanged together with the usernames john, ryan, and jennifer on 2025-01-31",
        expected_tool_calls=[
            ExpectedToolCall(
                func=get_messages_in_multi_person_dm_conversation_by_usernames,
                args={
                    "usernames": ["john", "ryan", "jennifer"],
                    "oldest_datetime": "2025-01-31 00:00:00",
                    "latest_datetime": "2025-01-31 23:59:59",
                },
            ),
        ],
        critics=[
            BinaryCritic(critic_field="usernames", weight=1 / 3),
            DatetimeCritic(
                critic_field="oldest_datetime", weight=1 / 3, max_difference=timedelta(minutes=2)
            ),
            DatetimeCritic(
                critic_field="latest_datetime", weight=1 / 3, max_difference=timedelta(minutes=2)
            ),
        ],
    )

    suite.add_case(
        name="Get conversation history oldest relative by username (2 days ago)",
        user_message="Get the messages I exchanged together with the usernames john, ryan, and jennifer starting 2 days ago",
        expected_tool_calls=[
            ExpectedToolCall(
                func=get_messages_in_multi_person_dm_conversation_by_usernames,
                args={
                    "usernames": ["john", "ryan", "jennifer"],
                    "oldest_relative": "02:00:00",
                },
            ),
        ],
        critics=[
            BinaryCritic(critic_field="usernames", weight=0.5),
            RelativeTimeBinaryCritic(critic_field="oldest_relative", weight=0.5),
        ],
    )

    return suite



import asyncio
from datetime import datetime, timezone
from typing import Annotated, cast

from arcade_tdk import ToolContext, tool
from arcade_tdk.auth import Slack
from arcade_tdk.errors import RetryableToolError, ToolExecutionError
from slack_sdk.errors import SlackApiError
from slack_sdk.web.async_client import AsyncWebClient

from arcade_slack.constants import MAX_PAGINATION_TIMEOUT_SECONDS
from arcade_slack.exceptions import (
    ItemNotFoundError,
    UsernameNotFoundError,
)
from arcade_slack.models import (
    ConversationType,
    SlackUserList,
)
from arcade_slack.tools.users import get_user_info_by_id, list_users
from arcade_slack.utils import (
    async_paginate,
    convert_conversation_type_to_slack_name,
    convert_datetime_to_unix_timestamp,
    convert_relative_datetime_to_unix_timestamp,
    enrich_message_datetime,
    extract_conversation_metadata,
    format_users,
    get_user_by_username,
    retrieve_conversations_by_user_ids,
)


@tool(
    requires_auth=Slack(
        scopes=[
            "chat:write",
            "im:write",
            "users.profile:read",
            "users:read",
        ],
    )
)
async def send_dm_to_user(
    context: ToolContext,
    user_name: Annotated[
        str,
        (
            "The Slack username of the person you want to message. "
            "Slack usernames are ALWAYS lowercase."
        ),
    ],
    message: Annotated[str, "The message you want to send"],
) -> Annotated[dict, "The response from the Slack API"]:
    """Send a direct message to a user in Slack."""

    token = (
        context.authorization.token if context.authorization and context.authorization.token else ""
    )
    slackClient = AsyncWebClient(token=token)

    try:
        # Step 1: Retrieve the user's Slack ID based on their username
        user_list_response = await slackClient.users_list()
        user_id = None
        for user in user_list_response["members"]:
            response_user_name = (
                "" if not isinstance(user.get("name"), str) else user["name"].lower()
            )
            if response_user_name == user_name.lower():
                user_id = user["id"]
                break

        if not user_id:
            raise RetryableToolError(
                "User not found",
                developer_message=f"User with username '{user_name}' not found.",
                additional_prompt_content=format_users(cast(SlackUserList, user_list_response)),
                retry_after_ms=500,  # Play nice with Slack API rate limits
            )

        # Step 2: Retrieve the DM channel ID with the user
        im_response = await slackClient.conversations_open(users=[user_id])
        dm_channel_id = im_response["channel"]["id"]

        # Step 3: Send the message as if it's from you (because we're using a user token)
        response = await slackClient.chat_postMessage(channel=dm_channel_id, text=message)

    except SlackApiError as e:
        error_message = e.response["error"] if "error" in e.response else str(e)
        raise ToolExecutionError(
            "Error sending message",
            developer_message=f"Slack API Error: {error_message}",
        )
    else:
        return {"response": response.data}


@tool(
    requires_auth=Slack(
        scopes=[
            "chat:write",
            "channels:read",
            "groups:read",
        ],
    )
)
async def send_message_to_channel(
    context: ToolContext,
    channel_name: Annotated[str, "The Slack channel name where you want to send the message. "],
    message: Annotated[str, "The message you want to send"],
) -> Annotated[dict, "The response from the Slack API"]:
    """Send a message to a channel in Slack."""

    try:
        slackClient = AsyncWebClient(
            token=context.authorization.token
            if context.authorization and context.authorization.token
            else ""
        )

        channel = await get_channel_metadata_by_name(context=context, channel_name=channel_name)
        channel_id = channel["id"]

        response = await slackClient.chat_postMessage(channel=channel_id, text=message)

    except SlackApiError as e:
        error_message = e.response["error"] if "error" in e.response else str(e)
        raise ToolExecutionError(
            "Error sending message",
            developer_message=f"Slack API Error: {error_message}",
        )
    else:
        return {"response": response.data}


@tool(
    requires_auth=Slack(
        scopes=[
            "channels:read",
            "groups:read",
            "im:read",
            "mpim:read",
            "users:read",
            "users:read.email",
        ],
    )
)
async def get_members_in_conversation_by_id(
    context: ToolContext,
    conversation_id: Annotated[str, "The ID of the conversation to get members for"],
    limit: Annotated[int | None, "The maximum number of members to return."] = None,
    next_cursor: Annotated[str | None, "The cursor to use for pagination."] = None,
) -> Annotated[dict, "Information about each member in the conversation"]:
    """Get the members of a conversation in Slack by the conversation's ID."""
    token = (
        context.authorization.token if context.authorization and context.authorization.token else ""
    )
    slackClient = AsyncWebClient(token=token)

    try:
        member_ids, next_cursor = await async_paginate(
            slackClient.conversations_members,
            "members",
            limit=limit,
            next_cursor=next_cursor,
            channel=conversation_id,
        )
    except SlackApiError as e:
        if e.response["error"] == "channel_not_found":
            conversations = await list_conversations_metadata(context)
            available_conversations = ", ".join(
                f"{conversation['id']} ({conversation['name']})"
                for conversation in conversations["conversations"]
            )

            raise RetryableToolError(
                "Conversation not found",
                developer_message=f"Conversation with ID '{conversation_id}' not found.",
                additional_prompt_content=f"Available conversations: {available_conversations}",
                retry_after_ms=500,
            )

    # Get the members' info
    # TODO: This will probably hit rate limits. We should probably call list_users() and
    # then filter the results instead.
    members = await asyncio.gather(*[
        get_user_info_by_id(context, member_id) for member_id in member_ids
    ])

    return {
        "members": [member for member in members if not member.get("is_bot")],
        "next_cursor": next_cursor,
    }


@tool(
    requires_auth=Slack(
        scopes=[
            "channels:read",
            "groups:read",
            "im:read",
            "mpim:read",
            "users:read",
            "users:read.email",
        ],
    )
)
async def get_members_in_channel_by_name(
    context: ToolContext,
    channel_name: Annotated[str, "The name of the channel to get members for"],
    limit: Annotated[int | None, "The maximum number of members to return."] = None,
    next_cursor: Annotated[str | None, "The cursor to use for pagination."] = None,
) -> Annotated[dict, "The channel members' IDs and Names"]:
    """Get the members of a conversation in Slack by the conversation's name."""
    channel = await get_channel_metadata_by_name(context=context, channel_name=channel_name)

    return await get_members_in_conversation_by_id(  # type: ignore[no-any-return]
        context=context,
        conversation_id=channel["id"],
        limit=limit,
        next_cursor=next_cursor,
    )


# TODO: make the function accept a current unix timestamp argument to allow testing without
# mocking. Have to wait until arcade.core.annotations.Inferrable is implemented, so that we
# can avoid exposing this arg to the LLM.
@tool(
    requires_auth=Slack(
        scopes=["channels:history", "groups:history", "im:history", "mpim:history"],
    )
)
async def get_messages_in_conversation_by_id(
    context: ToolContext,
    conversation_id: Annotated[str, "The ID of the conversation to get history for"],
    oldest_relative: Annotated[
        str | None,
        (
            "The oldest message to include in the results, specified as a time offset from the "
            "current time in the format 'DD:HH:MM'"
        ),
    ] = None,
    latest_relative: Annotated[
        str | None,
        (
            "The latest message to include in the results, specified as a time offset from the "
            "current time in the format 'DD:HH:MM'"
        ),
    ] = None,
    oldest_datetime: Annotated[
        str | None,
        (
            "The oldest message to include in the results, specified as a datetime object in the "
            "format 'YYYY-MM-DD HH:MM:SS'"
        ),
    ] = None,
    latest_datetime: Annotated[
        str | None,
        (
            "The latest message to include in the results, specified as a datetime object in the "
            "format 'YYYY-MM-DD HH:MM:SS'"
        ),
    ] = None,
    limit: Annotated[int | None, "The maximum number of messages to return."] = None,
    next_cursor: Annotated[str | None, "The cursor to use for pagination."] = None,
) -> Annotated[
    dict,
    (
        "The messages in a conversation and next cursor for paginating results (when "
        "there are additional messages to retrieve)."
    ),
]:
    """Get the messages in a conversation by the conversation's ID.

    A conversation can be a channel, a DM, or a group DM.

    To filter by an absolute datetime, use 'oldest_datetime' and/or 'latest_datetime'. If
    only 'oldest_datetime' is provided, it returns messages from the oldest_datetime to the
    current time. If only 'latest_datetime' is provided, it returns messages since the
    beginning of the conversation to the latest_datetime.

    To filter by a relative datetime (e.g. 3 days ago, 1 hour ago, etc.), use
    'oldest_relative' and/or 'latest_relative'. If only 'oldest_relative' is provided, it returns
    messages from the oldest_relative to the current time. If only 'latest_relative' is provided,
    it returns messages from the current time to the latest_relative.

    Do not provide both 'oldest_datetime' and 'oldest_relative' or both 'latest_datetime' and
    'latest_relative'.

    Leave all arguments with the default None to get messages without date/time filtering"""
    error_message = None
    if oldest_datetime and oldest_relative:
        error_message = "Cannot specify both 'oldest_datetime' and 'oldest_relative'."

    if latest_datetime and latest_relative:
        error_message = "Cannot specify both 'latest_datetime' and 'latest_relative'."

    if error_message:
        raise ToolExecutionError(error_message, developer_message=error_message)

    current_unix_timestamp = int(datetime.now(timezone.utc).timestamp())

    if latest_relative:
        latest_timestamp = convert_relative_datetime_to_unix_timestamp(
            latest_relative, current_unix_timestamp
        )
    elif latest_datetime:
        latest_timestamp = convert_datetime_to_unix_timestamp(latest_datetime)
    else:
        latest_timestamp = None

    if oldest_relative:
        oldest_timestamp = convert_relative_datetime_to_unix_timestamp(
            oldest_relative, current_unix_timestamp
        )
    elif oldest_datetime:
        oldest_timestamp = convert_datetime_to_unix_timestamp(oldest_datetime)
    else:
        oldest_timestamp = None

    token = (
        context.authorization.token if context.authorization and context.authorization.token else ""
    )
    slackClient = AsyncWebClient(token=token)

    datetime_args = {}
    if oldest_timestamp:
        datetime_args["oldest"] = oldest_timestamp
    if latest_timestamp:
        datetime_args["latest"] = latest_timestamp

    response, next_cursor = await async_paginate(
        slackClient.conversations_history,
        "messages",
        limit=limit,
        next_cursor=next_cursor,
        channel=conversation_id,
        include_all_metadata=True,
        inclusive=True,  # Include messages at the start and end of the time range
        **datetime_args,
    )

    messages = [enrich_message_datetime(message) for message in response]

    return {"messages": messages, "next_cursor": next_cursor}


# TODO: make the function accept a current unix timestamp argument to allow testing without
# mocking. Have to wait until arcade.core.annotations.Inferrable is implemented, so that we
# can avoid exposing this arg to the LLM.
@tool(
    requires_auth=Slack(
        scopes=[
            "channels:history",
            "channels:read",
            "groups:history",
            "groups:read",
            "im:history",
            "im:read",
            "mpim:history",
            "mpim:read",
        ],
    )
)
async def get_messages_in_channel_by_name(
    context: ToolContext,
    channel_name: Annotated[str, "The name of the channel"],
    oldest_relative: Annotated[
        str | None,
        (
            "The oldest message to include in the results, specified as a time offset from the "
            "current time in the format 'DD:HH:MM'"
        ),
    ] = None,
    latest_relative: Annotated[
        str | None,
        (
            "The latest message to include in the results, specified as a time offset from the "
            "current time in the format 'DD:HH:MM'"
        ),
    ] = None,
    oldest_datetime: Annotated[
        str | None,
        (
            "The oldest message to include in the results, specified as a datetime object in the "
            "format 'YYYY-MM-DD HH:MM:SS'"
        ),
    ] = None,
    latest_datetime: Annotated[
        str | None,
        (
            "The latest message to include in the results, specified as a datetime object in the "
            "format 'YYYY-MM-DD HH:MM:SS'"
        ),
    ] = None,
    limit: Annotated[int | None, "The maximum number of messages to return."] = None,
    next_cursor: Annotated[str | None, "The cursor to use for pagination."] = None,
) -> Annotated[
    dict,
    (
        "The messages in a channel and next cursor for paginating results (when "
        "there are additional messages to retrieve)."
    ),
]:
    """Get the messages in a channel by the channel's name.

    To filter messages by an absolute datetime, use 'oldest_datetime' and/or 'latest_datetime'. If
    only 'oldest_datetime' is provided, it will return messages from the oldest_datetime to the
    current time. If only 'latest_datetime' is provided, it will return messages since the
    beginning of the channel to the latest_datetime.

    To filter messages by a relative datetime (e.g. 3 days ago, 1 hour ago, etc.), use
    'oldest_relative' and/or 'latest_relative'. If only 'oldest_relative' is provided, it will
    return messages from the oldest_relative to the current time. If only 'latest_relative' is
    provided, it will return messages from the current time to the latest_relative.

    Do not provide both 'oldest_datetime' and 'oldest_relative' or both 'latest_datetime' and
    'latest_relative'.

    Leave all arguments with the default None to get messages without date/time filtering"""
    channel = await get_channel_metadata_by_name(context=context, channel_name=channel_name)

    return await get_messages_in_conversation_by_id(  # type: ignore[no-any-return]
        context=context,
        conversation_id=channel["id"],
        oldest_relative=oldest_relative,
        latest_relative=latest_relative,
        oldest_datetime=oldest_datetime,
        latest_datetime=latest_datetime,
        limit=limit,
        next_cursor=next_cursor,
    )


@tool(requires_auth=Slack(scopes=["im:history", "im:read"]))
async def get_messages_in_direct_message_conversation_by_username(
    context: ToolContext,
    username: Annotated[str, "The username of the user to get messages from"],
    oldest_relative: Annotated[
        str | None,
        (
            "The oldest message to include in the results, specified as a time offset from the "
            "current time in the format 'DD:HH:MM'"
        ),
    ] = None,
    latest_relative: Annotated[
        str | None,
        (
            "The latest message to include in the results, specified as a time offset from the "
            "current time in the format 'DD:HH:MM'"
        ),
    ] = None,
    oldest_datetime: Annotated[
        str | None,
        (
            "The oldest message to include in the results, specified as a datetime object in the "
            "format 'YYYY-MM-DD HH:MM:SS'"
        ),
    ] = None,
    latest_datetime: Annotated[
        str | None,
        (
            "The latest message to include in the results, specified as a datetime object in the "
            "format 'YYYY-MM-DD HH:MM:SS'"
        ),
    ] = None,
    limit: Annotated[int | None, "The maximum number of messages to return."] = None,
    next_cursor: Annotated[str | None, "The cursor to use for pagination."] = None,
) -> Annotated[
    dict,
    (
        "The messages in a direct message conversation and next cursor for paginating results "
        "when there are additional messages to retrieve."
    ),
]:
    """Get the messages in a direct conversation by the user's name.

    To filter messages by an absolute datetime, use 'oldest_datetime' and/or 'latest_datetime'. If
    only 'oldest_datetime' is provided, it will return messages from the oldest_datetime to the
    current time. If only 'latest_datetime' is provided, it will return messages since the
    beginning of the conversation to the latest_datetime.

    To filter messages by a relative datetime (e.g. 3 days ago, 1 hour ago, etc.), use
    'oldest_relative' and/or 'latest_relative'. If only 'oldest_relative' is provided, it will
    return messages from the oldest_relative to the current time. If only 'latest_relative' is
    provided, it will return messages from the current time to the latest_relative.

    Do not provide both 'oldest_datetime' and 'oldest_relative' or both 'latest_datetime' and
    'latest_relative'.

    Leave all arguments with the default None to get messages without date/time filtering"""
    direct_conversation = await get_direct_message_conversation_metadata_by_username(
        context=context, username=username
    )

    return await get_messages_in_conversation_by_id(  # type: ignore[no-any-return]
        context=context,
        conversation_id=direct_conversation["id"],
        oldest_relative=oldest_relative,
        latest_relative=latest_relative,
        oldest_datetime=oldest_datetime,
        latest_datetime=latest_datetime,
        limit=limit,
        next_cursor=next_cursor,
    )


@tool(requires_auth=Slack(scopes=["im:history", "im:read"]))
async def get_messages_in_multi_person_dm_conversation_by_usernames(
    context: ToolContext,
    usernames: Annotated[list[str], "The usernames of the users to get messages from"],
    oldest_relative: Annotated[
        str | None,
        (
            "The oldest message to include in the results, specified as a time offset from the "
            "current time in the format 'DD:HH:MM'"
        ),
    ] = None,
    latest_relative: Annotated[
        str | None,
        (
            "The latest message to include in the results, specified as a time offset from the "
            "current time in the format 'DD:HH:MM'"
        ),
    ] = None,
    oldest_datetime: Annotated[
        str | None,
        (
            "The oldest message to include in the results, specified as a datetime object in the "
            "format 'YYYY-MM-DD HH:MM:SS'"
        ),
    ] = None,
    latest_datetime: Annotated[
        str | None,
        (
            "The latest message to include in the results, specified as a datetime object in the "
            "format 'YYYY-MM-DD HH:MM:SS'"
        ),
    ] = None,
    limit: Annotated[int | None, "The maximum number of messages to return."] = None,
    next_cursor: Annotated[str | None, "The cursor to use for pagination."] = None,
) -> Annotated[
    dict,
    (
        "The messages in a multi-person direct message conversation and next cursor for "
        "paginating results (when there are additional messages to retrieve)."
    ),
]:
    """Get the messages in a multi-person direct message conversation by the usernames.

    To filter messages by an absolute datetime, use 'oldest_datetime' and/or 'latest_datetime'. If
    only 'oldest_datetime' is provided, it will return messages from the oldest_datetime to the
    current time. If only 'latest_datetime' is provided, it will return messages since the
    beginning of the conversation to the latest_datetime.

    To filter messages by a relative datetime (e.g. 3 days ago, 1 hour ago, etc.), use
    'oldest_relative' and/or 'latest_relative'. If only 'oldest_relative' is provided, it will
    return messages from the oldest_relative to the current time. If only 'latest_relative' is
    provided, it will return messages from the current time to the latest_relative.

    Do not provide both 'oldest_datetime' and 'oldest_relative' or both 'latest_datetime' and
    'latest_relative'.

    Leave all arguments with the default None to get messages without date/time filtering"""
    direct_conversation = await get_multi_person_dm_conversation_metadata_by_usernames(
        context=context, usernames=usernames
    )

    return await get_messages_in_conversation_by_id(  # type: ignore[no-any-return]
        context=context,
        conversation_id=direct_conversation["id"],
        oldest_relative=oldest_relative,
        latest_relative=latest_relative,
        oldest_datetime=oldest_datetime,
        latest_datetime=latest_datetime,
        limit=limit,
        next_cursor=next_cursor,
    )


@tool(
    requires_auth=Slack(
        scopes=["channels:read", "groups:read", "im:read", "mpim:read"],
    )
)
async def get_conversation_metadata_by_id(
    context: ToolContext,
    conversation_id: Annotated[str, "The ID of the conversation to get metadata for"],
) -> Annotated[dict, "The conversation metadata"]:
    """Get the metadata of a conversation in Slack searching by its ID.

    This tool does not return the messages in a conversation. To get the messages, use the
    `get_messages_in_conversation_by_id` tool."""
    token = (
        context.authorization.token if context.authorization and context.authorization.token else ""
    )
    slackClient = AsyncWebClient(token=token)

    try:
        response = await slackClient.conversations_info(
            channel=conversation_id,
            include_locale=True,
            include_num_members=True,
        )

    except SlackApiError as e:
        if e.response.get("error") == "channel_not_found":
            conversations = await list_conversations_metadata(context)
            available_conversations = ", ".join(
                f"{conversation['id']} ({conversation['name']})"
                for conversation in conversations["conversations"]
            )

            raise RetryableToolError(
                "Conversation not found",
                developer_message=f"Conversation with ID '{conversation_id}' not found.",
                additional_prompt_content=f"Available conversations: {available_conversations}",
                retry_after_ms=500,
            )

        raise

    return dict(**extract_conversation_metadata(response["channel"]))


@tool(requires_auth=Slack(scopes=["channels:read", "groups:read"]))
async def get_channel_metadata_by_name(
    context: ToolContext,
    channel_name: Annotated[str, "The name of the channel to get metadata for"],
    next_cursor: Annotated[
        str | None,
        "The cursor to use for pagination, if continuing from a previous search.",
    ] = None,
) -> Annotated[dict, "The channel metadata"]:
    """Get the metadata of a channel in Slack searching by its name.

    This tool does not return the messages in a channel. To get the messages, use the
    `get_messages_in_channel_by_name` tool."""
    channel_names: list[str] = []

    async def find_channel() -> dict:
        nonlocal channel_names, channel_name, next_cursor
        should_continue = True

        while should_continue:
            response = await list_conversations_metadata(
                context=context,
                conversation_types=[
                    ConversationType.PUBLIC_CHANNEL,
                    ConversationType.PRIVATE_CHANNEL,
                ],
                next_cursor=next_cursor,
            )
            next_cursor = response.get("next_cursor")

            for channel in response["conversations"]:
                response_channel_name = (
                    "" if not isinstance(channel.get("name"), str) else channel["name"].lower()
                )
                if response_channel_name == channel_name.lower():
                    return channel  # type: ignore[no-any-return]
                channel_names.append(channel["name"])

            if not next_cursor:
                should_continue = False

        raise ItemNotFoundError()

    try:
        return await asyncio.wait_for(find_channel(), timeout=MAX_PAGINATION_TIMEOUT_SECONDS)
    except ItemNotFoundError:
        raise RetryableToolError(
            "Channel not found",
            developer_message=f"Channel with name '{channel_name}' not found.",
            additional_prompt_content=f"Available channel names: {channel_names}",
            retry_after_ms=500,
        )
    except TimeoutError:
        raise RetryableToolError(
            "Channel not found, search timed out.",
            developer_message=(
                f"Channel with name '{channel_name}' not found. "
                f"Search timed out after {MAX_PAGINATION_TIMEOUT_SECONDS} seconds."
            ),
            additional_prompt_content=(
                f"Other channel names found are: {channel_names}. "
                "The list is potentially non-exhaustive, since the search process timed out. "
                f"Use the '{list_conversations_metadata.__tool_name__}' tool to get"
                "a comprehensive list of channels."
            ),
            retry_after_ms=500,
        )


@tool(requires_auth=Slack(scopes=["im:read"]))
async def get_direct_message_conversation_metadata_by_username(
    context: ToolContext,
    username: Annotated[str, "The username of the user/person to get messages with"],
    next_cursor: Annotated[
        str | None,
        "The cursor to use for pagination, if continuing from a previous search.",
    ] = None,
) -> Annotated[
    dict | None,
    "The direct message conversation metadata.",
]:
    """Get the metadata of a direct message conversation in Slack by the username.

    This tool does not return the messages in a conversation. To get the messages, use the
    `get_messages_in_direct_message_conversation_by_username` tool."""
    try:
        token = (
            context.authorization.token
            if context.authorization and context.authorization.token
            else ""
        )
        slack_client = AsyncWebClient(token=token)

        current_user, list_users_response = await asyncio.gather(
            slack_client.auth_test(), list_users(context)
        )

        other_user = get_user_by_username(username, list_users_response["users"])

        conversations_found = await retrieve_conversations_by_user_ids(
            list_conversations_func=list_conversations_metadata,
            get_members_in_conversation_func=get_members_in_conversation_by_id,
            context=context,
            conversation_types=[ConversationType.DIRECT_MESSAGE],
            user_ids=[current_user["user_id"], other_user["id"]],
            exact_match=True,
            limit=1,
            next_cursor=next_cursor,
        )

        return None if not conversations_found else conversations_found[0]

    except UsernameNotFoundError as e:
        raise RetryableToolError(
            f"Username '{e.username_not_found}' not found",
            developer_message=f"User with username '{e.username_not_found}' not found.",
            additional_prompt_content=f"Available users: {e.usernames_found}",
            retry_after_ms=500,
        )


@tool(requires_auth=Slack(scopes=["im:read"]))
async def get_multi_person_dm_conversation_metadata_by_usernames(
    context: ToolContext,
    usernames: Annotated[list[str], "The usernames of the users/people to get messages with"],
    next_cursor: Annotated[
        str | None,
        "The cursor to use for pagination, if continuing from a previous search.",
    ] = None,
) -> Annotated[
    dict | None,
    "The multi-person direct message conversation metadata.",
]:
    """Get the metadata of a multi-person direct message conversation in Slack by the usernames.

    This tool does not return the messages in a conversation. To get the messages, use the
    `get_messages_in_multi_person_dm_conversation_by_usernames` tool.
    """
    try:
        token = (
            context.authorization.token
            if context.authorization and context.authorization.token
            else ""
        )
        slack_client = AsyncWebClient(token=token)

        current_user, list_users_response = await asyncio.gather(
            slack_client.auth_test(), list_users(context)
        )

        other_users = [
            get_user_by_username(username, list_users_response["users"]) for username in usernames
        ]

        conversations_found = await retrieve_conversations_by_user_ids(
            list_conversations_func=list_conversations_metadata,
            get_members_in_conversation_func=get_members_in_conversation_by_id,
            context=context,
            conversation_types=[ConversationType.MULTI_PERSON_DIRECT_MESSAGE],
            user_ids=[
                current_user["user_id"],
                *[user["id"] for user in other_users if user["id"] != current_user["user_id"]],
            ],
            exact_match=True,
            limit=1,
            next_cursor=next_cursor,
        )

        if not conversations_found:
            raise RetryableToolError(
                "Conversation not found with the usernames provided",
                developer_message="Conversation not found with the usernames provided",
                retry_after_ms=500,
            )

        return conversations_found[0]

    except UsernameNotFoundError as e:
        raise RetryableToolError(
            f"Username '{e.username_not_found}' not found",
            developer_message=f"User with username '{e.username_not_found}' not found.",
            additional_prompt_content=f"Available users: {e.usernames_found}",
            retry_after_ms=500,
        )


@tool(
    requires_auth=Slack(
        scopes=["channels:read", "groups:read", "im:read", "mpim:read"],
    )
)
async def list_conversations_metadata(
    context: ToolContext,
    conversation_types: Annotated[
        list[ConversationType] | None,
        "The type(s) of conversations to list. Defaults to all types.",
    ] = None,
    limit: Annotated[int | None, "The maximum number of conversations to list."] = None,
    next_cursor: Annotated[str | None, "The cursor to use for pagination."] = None,
) -> Annotated[
    dict,
    (
        "The conversations metadata list and a pagination 'next_cursor', if there are more "
        "conversations to retrieve."
    ),
]:
    """
    List metadata for Slack conversations (channels and/or direct messages) that the user
    is a member of.
    """
    if isinstance(conversation_types, ConversationType):
        conversation_types = [conversation_types]

    conversation_types_filter = ",".join(
        convert_conversation_type_to_slack_name(conv_type).value
        for conv_type in conversation_types or ConversationType
    )

    token = (
        context.authorization.token if context.authorization and context.authorization.token else ""
    )
    slackClient = AsyncWebClient(token=token)

    results, next_cursor = await async_paginate(
        slackClient.conversations_list,
        "channels",
        limit=limit,
        next_cursor=next_cursor,
        types=conversation_types_filter,
        exclude_archived=True,
    )

    return {
        "conversations": [
            dict(**extract_conversation_metadata(conversation))
            for conversation in results
            if conversation.get("is_im") or conversation.get("is_member")
        ],
        "next_cursor": next_cursor,
    }


@tool(
    requires_auth=Slack(
        scopes=["channels:read"],
    )
)
async def list_public_channels_metadata(
    context: ToolContext,
    limit: Annotated[int | None, "The maximum number of channels to list."] = None,
) -> Annotated[dict, "The public channels"]:
    """List metadata for public channels in Slack that the user is a member of."""

    return await list_conversations_metadata(  # type: ignore[no-any-return]
        context,
        conversation_types=[ConversationType.PUBLIC_CHANNEL],
        limit=limit,
    )


@tool(
    requires_auth=Slack(
        scopes=["groups:read"],
    )
)
async def list_private_channels_metadata(
    context: ToolContext,
    limit: Annotated[int | None, "The maximum number of channels to list."] = None,
) -> Annotated[dict, "The private channels"]:
    """List metadata for private channels in Slack that the user is a member of."""

    return await list_conversations_metadata(  # type: ignore[no-any-return]
        context,
        conversation_types=[ConversationType.PRIVATE_CHANNEL],
        limit=limit,
    )


@tool(
    requires_auth=Slack(
        scopes=["mpim:read"],
    )
)
async def list_group_direct_message_conversations_metadata(
    context: ToolContext,
    limit: Annotated[int | None, "The maximum number of conversations to list."] = None,
) -> Annotated[dict, "The group direct message conversations metadata"]:
    """List metadata for group direct message conversations that the user is a member of."""

    return await list_conversations_metadata(  # type: ignore[no-any-return]
        context,
        conversation_types=[ConversationType.MULTI_PERSON_DIRECT_MESSAGE],
        limit=limit,
    )


# Note: Bots are included in the results.
# Note: Direct messages with no conversation history are included in the results.
@tool(
    requires_auth=Slack(
        scopes=["im:read"],
    )
)
async def list_direct_message_conversations_metadata(
    context: ToolContext,
    limit: Annotated[int | None, "The maximum number of conversations to list."] = None,
) -> Annotated[dict, "The direct message conversations metadata"]:
    """List metadata for direct message conversations in Slack that the user is a member of."""

    response = await list_conversations_metadata(
        context,
        conversation_types=[ConversationType.DIRECT_MESSAGE],
        limit=limit,
    )

    return response  # type: ignore[no-any-return]



import asyncio
from unittest.mock import AsyncMock, MagicMock, call, patch

import pytest
from arcade_tdk import ToolContext
from slack_sdk.errors import SlackApiError
from slack_sdk.web.async_client import AsyncWebClient

from arcade_slack.exceptions import PaginationTimeoutError
from arcade_slack.models import (
    ConversationType,
    ConversationTypeSlackName,
)
from arcade_slack.tools.chat import (
    get_members_in_conversation_by_id,
    list_conversations_metadata,
)
from arcade_slack.utils import (
    async_paginate,
    convert_conversation_type_to_slack_name,
    filter_conversations_by_user_ids,
    retrieve_conversations_by_user_ids,
)


@pytest.mark.asyncio
async def test_async_paginate():
    mock_slack_client = AsyncMock()
    mock_slack_client.conversations_list.return_value = {
        "ok": True,
        "channels": [{"id": "123"}],
        "response_metadata": {"next_cursor": None},
    }

    results, next_cursor = await async_paginate(
        func=mock_slack_client.conversations_list,
        response_key="channels",
    )

    assert results == [{"id": "123"}]
    assert next_cursor is None


@pytest.mark.asyncio
async def test_async_paginate_with_response_error():
    mock_slack_client = AsyncMock()
    mock_slack_client.conversations_list.side_effect = SlackApiError(
        message="slack_error",
        response={"ok": False, "error": "slack_error"},
    )

    with pytest.raises(SlackApiError) as e:
        await async_paginate(
            func=mock_slack_client.conversations_list,
            response_key="channels",
        )
        assert str(e.value) == "slack_error"


@pytest.mark.asyncio
async def test_async_paginate_with_custom_pagination_args():
    mock_slack_client = AsyncMock()
    mock_slack_client.conversations_list.return_value = {
        "ok": True,
        "channels": [{"id": "123"}],
        "response_metadata": {"next_cursor": "456"},
    }

    results, next_cursor = await async_paginate(
        func=mock_slack_client.conversations_list,
        response_key="channels",
        limit=1,
        next_cursor="123",
        hello="world",
    )

    assert results == [{"id": "123"}]
    assert next_cursor == "456"

    mock_slack_client.conversations_list.assert_called_once_with(
        hello="world",
        limit=1,
        cursor="123",
    )


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "test_limit, last_next_cursor, last_expected_limit",
    [(5, "cursor3", 1), (None, None, 2)],
)
async def test_async_paginate_large_limit(test_limit, last_next_cursor, last_expected_limit):
    mock_slack_client = AsyncMock(spec=AsyncWebClient)
    mock_slack_client.conversations_list.side_effect = [
        {
            "ok": True,
            "channels": [{"id": "channel1"}, {"id": "channel2"}],
            "response_metadata": {"next_cursor": "cursor1"},
        },
        {
            "ok": True,
            "channels": [{"id": "channel3"}, {"id": "channel4"}],
            "response_metadata": {"next_cursor": "cursor2"},
        },
        {
            "ok": True,
            "channels": [{"id": "channel5"}],
            "response_metadata": {"next_cursor": last_next_cursor},
        },
    ]

    with patch("arcade_slack.utils.MAX_PAGINATION_SIZE_LIMIT", 2):
        results, next_cursor = await async_paginate(
            func=mock_slack_client.conversations_list,
            response_key="channels",
            limit=test_limit,
            hello="world",
        )

    assert results == [
        {"id": "channel1"},
        {"id": "channel2"},
        {"id": "channel3"},
        {"id": "channel4"},
        {"id": "channel5"},
    ]
    assert next_cursor == last_next_cursor
    assert mock_slack_client.conversations_list.call_count == 3
    mock_slack_client.conversations_list.assert_has_calls([
        call(hello="world", limit=2, cursor=None),
        call(hello="world", limit=2, cursor="cursor1"),
        call(hello="world", limit=last_expected_limit, cursor="cursor2"),
    ])


@pytest.mark.asyncio
async def test_async_paginate_large_limit_with_response_error():
    mock_slack_client = AsyncMock()
    mock_slack_client.conversations_list.side_effect = [
        {
            "ok": True,
            "channels": [{"id": "channel1"}, {"id": "channel2"}],
            "response_metadata": {"next_cursor": "cursor1"},
        },
        SlackApiError(message="slack_error", response={"ok": False, "error": "slack_error"}),
        {
            "ok": True,
            "channels": [{"id": "channel5"}],
            "response_metadata": {"next_cursor": "cursor3"},
        },
    ]

    with (
        patch("arcade_slack.utils.MAX_PAGINATION_SIZE_LIMIT", 2),
        pytest.raises(SlackApiError) as e,
    ):
        await async_paginate(
            func=mock_slack_client.conversations_list,
            response_key="channels",
            limit=5,
            hello="world",
        )
        assert str(e.value) == "slack_error"

    assert mock_slack_client.conversations_list.call_count == 2
    mock_slack_client.conversations_list.assert_has_calls([
        call(hello="world", limit=2, cursor=None),
        call(hello="world", limit=2, cursor="cursor1"),
    ])


@pytest.mark.asyncio
async def test_async_paginate_with_timeout():
    # Mock Slack client
    mock_slack_client = AsyncMock()

    # Simulate a network delay by making the mock function sleep
    async def mock_conversations_list(*args, **kwargs):
        await asyncio.sleep(1)  # Sleep for 1 second to simulate delay
        return {
            "ok": True,
            "channels": [{"id": "123"}],
            "response_metadata": {"next_cursor": None},
        }

    mock_slack_client.conversations_list.side_effect = mock_conversations_list

    # Set a low timeout to trigger the timeout error quickly during the test
    max_pagination_timeout_seconds = 0.1  # 100 milliseconds

    with pytest.raises(PaginationTimeoutError) as exc_info:
        await async_paginate(
            func=mock_slack_client.conversations_list,
            response_key="channels",
            max_pagination_timeout_seconds=max_pagination_timeout_seconds,
        )

    assert (
        str(exc_info.value)
        == f"The pagination process timed out after {max_pagination_timeout_seconds} seconds."
    )


def test_filter_conversations_by_user_ids():
    conversations = [
        {"id": "123", "members": [{"id": "user1"}, {"id": "user2"}, {"id": "user3"}]},
        {"id": "456", "members": [{"id": "user2"}, {"id": "user3"}]},
    ]
    response = filter_conversations_by_user_ids(
        conversations=conversations,
        user_ids=["user1", "user2"],
        exact_match=False,
    )
    assert response == [
        {"id": "123", "members": [{"id": "user1"}, {"id": "user2"}, {"id": "user3"}]},
    ]


def test_filter_conversations_by_user_ids_empty_response():
    conversations = [
        {"id": "123", "members": [{"id": "user1"}, {"id": "user3"}, {"id": "user4"}]},
        {"id": "456", "members": [{"id": "user2"}, {"id": "user3"}, {"id": "user4"}]},
    ]
    response = filter_conversations_by_user_ids(
        conversations=conversations,
        user_ids=["user1", "user2"],
        exact_match=False,
    )
    assert response == []


def test_filter_conversations_by_user_ids_multiple_matches():
    conversations = [
        {"id": "123", "members": [{"id": "user1"}, {"id": "user2"}, {"id": "user3"}]},
        {"id": "456", "members": [{"id": "user2"}, {"id": "user3"}]},
        {
            "id": "789",
            "members": [{"id": "user4"}, {"id": "user1"}, {"id": "user2"}, {"id": "user3"}],
        },
    ]
    response = filter_conversations_by_user_ids(
        conversations=conversations,
        user_ids=["user1", "user2"],
        exact_match=False,
    )
    assert response == [
        {"id": "123", "members": [{"id": "user1"}, {"id": "user2"}, {"id": "user3"}]},
        {
            "id": "789",
            "members": [{"id": "user4"}, {"id": "user1"}, {"id": "user2"}, {"id": "user3"}],
        },
    ]


def test_filter_conversations_by_user_ids_exact_match():
    conversations = [
        {"id": "123", "members": [{"id": "user1"}, {"id": "user2"}]},
        {"id": "456", "members": [{"id": "user2"}, {"id": "user3"}]},
    ]
    response = filter_conversations_by_user_ids(
        conversations=conversations,
        user_ids=["user1", "user2"],
        exact_match=True,
    )
    assert response == [{"id": "123", "members": [{"id": "user1"}, {"id": "user2"}]}]


def test_filter_conversations_by_user_ids_exact_match_empty_response():
    conversations = [
        {"id": "123", "members": [{"id": "user1"}, {"id": "user2"}, {"id": "user3"}]},
        {"id": "456", "members": [{"id": "user2"}, {"id": "user3"}]},
    ]
    response = filter_conversations_by_user_ids(
        conversations=conversations,
        user_ids=["user1", "user2"],
        exact_match=True,
    )
    assert response == []


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "search_user_ids, conversation_types, exact_match, limit, expected_conversation_ids",
    [
        (["U1", "U2"], [ConversationType.DIRECT_MESSAGE], False, 1, ["C1"]),
        (["U1", "U2"], [ConversationType.DIRECT_MESSAGE], True, 1, ["C1"]),
        (["U1", "U2", "U3"], [ConversationType.DIRECT_MESSAGE], False, 1, []),
        (
            ["U1", "U2"],
            [ConversationType.DIRECT_MESSAGE, ConversationType.PUBLIC_CHANNEL],
            False,
            10,
            ["C1", "C3", "C4"],
        ),
        (
            ["U1", "U2"],
            [ConversationType.DIRECT_MESSAGE, ConversationType.PUBLIC_CHANNEL],
            True,
            10,
            ["C1", "C3"],
        ),
    ],
)
async def test_retrieve_conversations_by_user_ids(
    mock_chat_slack_client,
    mock_users_slack_client,
    search_user_ids,
    conversation_types,
    exact_match,
    limit,
    expected_conversation_ids,
):
    context = MagicMock(spec=ToolContext)
    context.authorization = MagicMock()
    context.authorization.token = MagicMock()

    conversation_types_slack_name_str = [
        convert_conversation_type_to_slack_name(conv_type).value
        for conv_type in conversation_types or ConversationType
    ]

    conversations = [
        {
            "conversation": {
                "id": "C1",
                "type": ConversationTypeSlackName.IM.value,
                "name": "im-1",
                "is_channel": False,
                "is_im": True,
                "is_member": True,
            },
            "members": {
                "ok": True,
                "members": ["U1", "U2"],
                "response_metadata": {"next_cursor": None},
            },
            "users": [
                {"ok": True, "user": {"id": "U1", "team_id": "T123", "name": "user1"}},
                {"ok": True, "user": {"id": "U2", "team_id": "T123", "name": "user2"}},
            ],
        },
        {
            "conversation": {
                "id": "C2",
                "type": ConversationTypeSlackName.IM.value,
                "name": "im-2",
                "is_channel": False,
                "is_im": True,
                "is_member": True,
            },
            "members": {
                "ok": True,
                "members": ["U2", "U3"],
                "response_metadata": {"next_cursor": None},
            },
            "users": [
                {"ok": True, "user": {"id": "U2", "team_id": "T123", "name": "user2"}},
                {"ok": True, "user": {"id": "U3", "team_id": "T123", "name": "user3"}},
            ],
        },
        {
            "conversation": {
                "id": "C3",
                "type": ConversationTypeSlackName.PUBLIC_CHANNEL.value,
                "name": "general",
                "is_channel": True,
                "is_im": False,
                "is_member": True,
            },
            "members": {
                "ok": True,
                "members": ["U1", "U2"],
                "response_metadata": {"next_cursor": None},
            },
            "users": [
                {"ok": True, "user": {"id": "U1", "team_id": "T123", "name": "user1"}},
                {"ok": True, "user": {"id": "U2", "team_id": "T123", "name": "user2"}},
            ],
        },
        {
            "conversation": {
                "id": "C4",
                "type": ConversationTypeSlackName.PUBLIC_CHANNEL.value,
                "name": "random",
                "is_channel": True,
                "is_im": False,
                "is_member": True,
            },
            "members": {
                "ok": True,
                "members": ["U1", "U2", "U3", "U4"],
                "response_metadata": {"next_cursor": None},
            },
            "users": [
                {"ok": True, "user": {"id": "U1", "team_id": "T123", "name": "user1"}},
                {"ok": True, "user": {"id": "U2", "team_id": "T123", "name": "user2"}},
                {"ok": True, "user": {"id": "U3", "team_id": "T123", "name": "user3"}},
                {"ok": True, "user": {"id": "U4", "team_id": "T123", "name": "user4"}},
            ],
        },
    ]

    conversations_listed = [
        conversation
        for conversation in conversations
        if conversation["conversation"]["type"] in conversation_types_slack_name_str
    ]

    mock_chat_slack_client.conversations_list.return_value = {
        "ok": True,
        "channels": [conversation["conversation"] for conversation in conversations_listed],
        "response_metadata": {"next_cursor": None},
    }

    mock_chat_slack_client.conversations_members.side_effect = [
        conversation["members"] for conversation in conversations_listed
    ]

    mock_users_slack_client.users_info.side_effect = [
        user for conversation in conversations_listed for user in conversation["users"]
    ]

    conversations_found = await retrieve_conversations_by_user_ids(
        list_conversations_func=list_conversations_metadata,
        get_members_in_conversation_func=get_members_in_conversation_by_id,
        context=context,
        conversation_types=conversation_types,
        user_ids=search_user_ids,
        exact_match=exact_match,
        limit=limit,
        next_cursor=None,
    )

    assert [conversation["id"] for conversation in conversations_found] == expected_conversation_ids


@pytest.mark.asyncio
@pytest.mark.parametrize(
    (
        "search_user_ids, conversation_types, exact_match, limit, "
        "expected_conversation_ids, expected_conversations_list_calls"
    ),
    [
        (
            ["U1", "U2", "U3"],
            [ConversationType.MULTI_PERSON_DIRECT_MESSAGE],
            False,
            None,
            ["C1", "C3"],
            2,
        ),
        (
            ["U1", "U2", "U3"],
            [ConversationType.MULTI_PERSON_DIRECT_MESSAGE],
            True,
            None,
            ["C1"],
            2,
        ),
        (["U1", "U2", "U99"], [ConversationType.MULTI_PERSON_DIRECT_MESSAGE], False, None, [], 2),
        (
            ["U1", "U2"],
            [ConversationType.MULTI_PERSON_DIRECT_MESSAGE, ConversationType.PUBLIC_CHANNEL],
            False,
            None,
            ["C1", "C3", "C4", "C6"],
            2,
        ),
        (
            ["U1", "U2"],
            [ConversationType.MULTI_PERSON_DIRECT_MESSAGE, ConversationType.PUBLIC_CHANNEL],
            False,
            1,
            ["C1"],
            2,
        ),
        (
            ["U1", "U2"],
            [ConversationType.MULTI_PERSON_DIRECT_MESSAGE, ConversationType.PUBLIC_CHANNEL],
            False,
            3,
            ["C1", "C3", "C4"],
            2,
        ),
        (
            ["U1", "U2"],
            [ConversationType.MULTI_PERSON_DIRECT_MESSAGE, ConversationType.PUBLIC_CHANNEL],
            True,
            None,
            ["C4"],
            2,
        ),
    ],
)
async def test_retrieve_conversations_by_user_ids_with_pagination(
    mock_chat_slack_client,
    mock_users_slack_client,
    search_user_ids,
    conversation_types,
    exact_match,
    limit,
    expected_conversation_ids,
    expected_conversations_list_calls,
):
    context = MagicMock(spec=ToolContext)
    context.authorization = MagicMock()
    context.authorization.token = MagicMock()

    conversation_types_slack_name_str = [
        convert_conversation_type_to_slack_name(conv_type).value
        for conv_type in conversation_types or ConversationType
    ]

    conversations = [
        {
            "conversation": {
                "id": "C1",
                "type": ConversationTypeSlackName.MPIM.value,
                "name": "mpim-1",
                "is_channel": False,
                "is_im": False,
                "is_mpim": True,
                "is_member": True,
            },
            "members": {
                "ok": True,
                "members": ["U1", "U2", "U3"],
                "response_metadata": {"next_cursor": None},
            },
            "users": [
                {"ok": True, "user": {"id": "U1", "team_id": "T123", "name": "user1"}},
                {"ok": True, "user": {"id": "U2", "team_id": "T123", "name": "user2"}},
                {"ok": True, "user": {"id": "U3", "team_id": "T123", "name": "user3"}},
            ],
        },
        {
            "conversation": {
                "id": "C2",
                "type": ConversationTypeSlackName.MPIM.value,
                "name": "mpim-2",
                "is_channel": False,
                "is_im": False,
                "is_mpim": True,
                "is_member": True,
            },
            "members": {
                "ok": True,
                "members": ["U2", "U3"],
                "response_metadata": {"next_cursor": None},
            },
            "users": [
                {"ok": True, "user": {"id": "U2", "team_id": "T123", "name": "user2"}},
                {"ok": True, "user": {"id": "U3", "team_id": "T123", "name": "user3"}},
            ],
        },
        {
            "conversation": {
                "id": "C3",
                "type": ConversationTypeSlackName.MPIM.value,
                "name": "mpim-3",
                "is_channel": False,
                "is_im": False,
                "is_mpim": True,
                "is_member": True,
            },
            "members": {
                "ok": True,
                "members": ["U1", "U2", "U3", "U4"],
                "response_metadata": {"next_cursor": None},
            },
            "users": [
                {"ok": True, "user": {"id": "U1", "team_id": "T123", "name": "user1"}},
                {"ok": True, "user": {"id": "U2", "team_id": "T123", "name": "user2"}},
                {"ok": True, "user": {"id": "U3", "team_id": "T123", "name": "user3"}},
                {"ok": True, "user": {"id": "U4", "team_id": "T123", "name": "user4"}},
            ],
        },
        {
            "conversation": {
                "id": "C4",
                "type": ConversationTypeSlackName.PUBLIC_CHANNEL.value,
                "name": "channel-4",
                "is_channel": True,
                "is_im": False,
                "is_member": True,
            },
            "members": {
                "ok": True,
                "members": ["U1", "U2"],
                "response_metadata": {"next_cursor": None},
            },
            "users": [
                {"ok": True, "user": {"id": "U1", "team_id": "T123", "name": "user1"}},
                {"ok": True, "user": {"id": "U2", "team_id": "T123", "name": "user2"}},
            ],
        },
        {
            "conversation": {
                "id": "C5",
                "type": ConversationTypeSlackName.PUBLIC_CHANNEL.value,
                "name": "channel-5",
                "is_channel": True,
                "is_im": False,
                "is_member": True,
            },
            "members": {
                "ok": True,
                "members": ["U2", "U3", "U4"],
                "response_metadata": {"next_cursor": None},
            },
            "users": [
                {"ok": True, "user": {"id": "U2", "team_id": "T123", "name": "user2"}},
                {"ok": True, "user": {"id": "U3", "team_id": "T123", "name": "user3"}},
                {"ok": True, "user": {"id": "U4", "team_id": "T123", "name": "user4"}},
            ],
        },
        {
            "conversation": {
                "id": "C6",
                "type": ConversationTypeSlackName.PUBLIC_CHANNEL.value,
                "name": "channel-6",
                "is_channel": True,
                "is_im": False,
                "is_member": True,
            },
            "members": {
                "ok": True,
                "members": ["U1", "U2", "U3", "U4"],
                "response_metadata": {"next_cursor": None},
            },
            "users": [
                {"ok": True, "user": {"id": "U1", "team_id": "T123", "name": "user1"}},
                {"ok": True, "user": {"id": "U2", "team_id": "T123", "name": "user2"}},
                {"ok": True, "user": {"id": "U3", "team_id": "T123", "name": "user3"}},
                {"ok": True, "user": {"id": "U4", "team_id": "T123", "name": "user4"}},
            ],
        },
    ]

    conversations_listed = [
        conversation
        for conversation in conversations
        if conversation["conversation"]["type"] in conversation_types_slack_name_str
    ]

    split_size = len(conversations_listed) // 2

    conversations_listed_1 = conversations_listed[:split_size]
    conversations_listed_2 = conversations_listed[split_size:]

    mock_chat_slack_client.conversations_list.side_effect = [
        {
            "ok": True,
            "channels": [conversation["conversation"] for conversation in conversations_listed_1],
            "response_metadata": {"next_cursor": "cursor_1"},
        },
        {
            "ok": True,
            "channels": [conversation["conversation"] for conversation in conversations_listed_2],
            "response_metadata": {"next_cursor": None},
        },
    ]

    mock_chat_slack_client.conversations_members.side_effect = [
        conversation["members"] for conversation in conversations_listed
    ]

    mock_users_slack_client.users_info.side_effect = [
        user for conversation in conversations_listed for user in conversation["users"]
    ]

    conversations_found = await retrieve_conversations_by_user_ids(
        list_conversations_func=list_conversations_metadata,
        get_members_in_conversation_func=get_members_in_conversation_by_id,
        context=context,
        conversation_types=conversation_types,
        user_ids=search_user_ids,
        exact_match=exact_match,
        limit=limit,
        next_cursor=None,
    )

    assert [conversation["id"] for conversation in conversations_found] == expected_conversation_ids
    assert mock_chat_slack_client.conversations_list.call_count == expected_conversations_list_calls
